"""
File Purpose: subsampling "main" dimensions (and snapshot dimension)

Produce a subsampled version of the data where
some of the variables have been subsampled on main dimensions.
"""
import json
import os
import re

import numpy as np

from ..errors import (
    InputError, InputMissingError, InputConflictError,
    SubsamplingError, SubsamplingFormatError, SubsamplingNotFoundError,
)
from ..tools import (
    simple_property, alias,
    UNSET,
    is_iterable,
    ProgressUpdater,
    code_snapshot_info, nbytes_path,
)


''' --------------------- SubsamplingInfo and methods --------------------- '''

class SubsamplingInfo():
    '''information about subsampling
    
    json: dict of subsampling info: {
        'forall': {       # optional; if not provided use vals from byvars.
            'snaps': [mode, info],                   # optional, can provide here or in byvars but not both.
            'array_dims': [list of array dim names in order],      # provide here or in byvars but not both.
            },
        'byvars': [       # optional; if not provided do snaps slicing only
            {
            'snaps': [mode, info],                   # optional, can provide here or in forall but not both.
            'array_dims': [list of array dim names in order],      # provide here or in forall but not both.
            'exact': {   # optional
                var1: [mode1, info1],
                var2: [mode2, info2],  # optional
                ...,
                },
            'regex': {   # optional
                re_var3: [mode3, info3],
                ...,
                },
            },
            ...,  # optionally can provide more dicts like above
            ]
        }
        where:
            snaps = how to subsample snapshots (use dim='snap' for these). Example:
                'snaps': ['slice', {'snap': [start, stop, step]}]
            exact = dict of exact names; subsampling for var with this name
            regex = dict of regex patterns; subsampling for var if re.fullmatch(pattern, var)
            mode = mode for subsampling. e.g. 'slice'
            info = info for subsampling, format depends on mode.
                e.g. for 'slice', info = dict of {dim: [start, stop, step]}.
        Note: if multiple matches for any var, always use the earliest one;
            i.e. earliest dict in list, and always exact match before regex match.
    meta: dict, None, or UNSET
        metadata about the subsampling info. Must be in json format, if provided.
        None --> no metadata
        UNSET --> use code_snapshot_info() to get metadata when this object is created.
    _json_indent: int
        number of spaces to indent when saving json file.
    _init_appliers: bool
        whether to self.init_appliers() during __init__.
        Usually should be True, but if debugging might want to use False.
    '''
    # subsampling_path_manager_cls = SubsamplingInfoPathManager  # <-- defined lower in this file!

    def __init__(self, json, *, meta=UNSET, _json_indent=4, _init_appliers=True):
        self.json = json
        self._json_indent = _json_indent
        self.meta = meta
        self.init_meta()
        self.init_checks()
        if _init_appliers:
            self.init_appliers()

    # # # METHODS -- IMPLEMENTED HERE # # #
    @classmethod
    def implied_from(cls, json_or_filename_or_info, **kw_cls):
        '''return cls object from json, filename, or SubsamplingInfo object.'''
        x = json_or_filename_or_info
        if isinstance(x, SubsamplingInfo):
            info = x
            if isinstance(info, cls):
                return info
            else:
                kw.setdefault('_json_indent', info._json_indent)
                return cls(info.json, **kw_cls)
        elif isinstance(x, str):
            filename = x
            return cls.load(filename, **kw_cls)
        elif isinstance(x, dict):
            json = x
            return cls(json, **kw_cls)
        else:
            raise InputError(f'cannot create {cls.__name__} from {type(x)}.')

    def init_appliers(self):
        '''initialize all subsampling appliers for this SubsamplingInfo object.'''
        aa = {}
        # handle the forall stuff
        forall = self.json.get('forall', {})
        forall_array_dims = forall.get('array_dims', None)
        forall_snap_applier = None
        if self.has_forall_snaps():
            aa['forall'] = {}
            mode, info = forall['snaps']
            snap_applier = self.MODES[mode](info, array_dims=['snap'])
            aa['forall']['snaps'] = snap_applier
        # handle the byvars stuff
        aa['byvars'] = []
        for vdict in self.json.get('byvars', []):
            # handle the possibly-forall stuff
            array_dims = forall_array_dims
            if array_dims is None:
                array_dims = vdict.get('array_dims', [])
            snap_applier = forall_snap_applier
            if snap_applier is None:
                snapstuff = vdict.get('snaps', None)
                if snapstuff is not None:
                    mode, info = snapstuff
                    snap_applier = self.MODES[mode](info, array_dims=['snap'])
            # handle the definitely-byvars stuff
            adict = dict()
            for exact in [ex_ for ex_ in ['exact', 'regex'] if ex_ in vdict]:
                dd = dict()
                for var, [mode, info] in vdict[exact].items():
                    subsampler_cls = self.MODES[mode]
                    dd[var] = subsampler_cls(info, array_dims=array_dims, snap_applier=snap_applier)
                adict[exact] = dd
            aa['byvars'].append(adict)
        self.appliers = aa

    def init_checks(self):
        '''do some sanity checks about this SubsamplingInfo object.
        Raise SubsamplingFormatError if any checks fail.
        Here, checks:
            - uniqueness of var keys: expect to not have exact same var key appear in multiple places.
            - snaps can appear in forall or byvars, but not both.
            - array_dims can appear in forall or byvars, but not both.
            - array_dims must be provided (possibly via forall) for any byvars dict with at least 1 var.
        '''
        # uniqueness of var keys:
        all_vars = set()
        for vdict in self.json.get('byvars', []):
            for exact in [ex_ for ex_ in ['exact', 'regex'] if ex_ in vdict]:
                for var in vdict[exact]:
                    if var in all_vars:
                        raise SubsamplingFormatError(f'var {var!r} appears in multiple places.')
                    all_vars.add(var)
        # snaps can appear in globals or in individual dicts, but not both.
        if 'snaps' in self.json.get('forall', {}):
            for vdict in self.json.get('byvars', []):
                if 'snaps' in vdict:
                    raise SubsamplingFormatError('snaps can appear in globals or in individual dicts, but not both.')
        # array_dims can appear in globals or in individual dicts, but not both.
        if 'array_dims' in self.json.get('forall', {}):
            for vdict in self.json.get('byvars', []):
                if 'array_dims' in vdict:
                    raise SubsamplingFormatError('array_dims can appear in globals or in individual dicts, but not both.')
        # array_dims must be provided for any byvars dict with at least 1 var.
        if 'array_dims' not in self.json.get('forall', {}):
            for vdict in self.json.get('byvars', []):
                if 'array_dims' not in vdict:
                    for exact in [ex_ for ex_ in ['exact', 'regex'] if ex_ in vdict]:
                        if len(vdict[exact]) > 0:
                            errmsg = 'array_dims must be provided via forall or in each byvars dict with at least 1 var'
                            raise SubsamplingFormatError(errmsg)

    # # # META UPDATES # # #
    def init_meta(self):
        '''initialize self.meta. None remains None. UNSET becomes code_snapshot_info().'''
        if self.meta is UNSET:
            self.meta = code_snapshot_info()

    def update_meta(self, d):
        '''self.meta.update(d). If self.meta is None, keep it as None.'''
        # [TODO] more sophisticated update options?
        if self.meta is not None:
            self.meta.update(d)

    # # # SNAPS BASIC INFO # # #
    def has_forall_snaps(self):
        '''returns True if 'snaps' in 'forall', else False.'''
        return 'snaps' in self.json.get('forall', {})

    def has_byvars_snaps(self):
        '''return True if 'snaps' in any byvars, else False.'''
        return any('snaps' in vdict for vdict in self.json.get('byvars', []))

    # # # INFO / APPLIER FOR A SINGLE VAR # # #
    def get_applier(self, var, *, missing_ok=True):
        '''returns a SubsamplingMode instance appropriate for var based on self.
        Assumes self.init_appliers() was run and is up-to-date.

        var: str
            variable name to get applier for, or string 'snaps'.
            'snaps' --> get 'snaps' applier from forall, else crash.
            else --> get first match from byvars.
        missing_ok: bool
            When var has no match in self:
                if missing_ok: return SubsamplingIdentity instance.
                otherwise: raise SubsamplingNotFoundError.
        '''
        aa = self.appliers
        if var == 'snaps':
            if self.has_forall_snaps():
                return aa['forall']['snaps']
            else:
                raise SubsamplingNotFoundError('no "forall" subsampling applier for snaps.')
        # else
        for adict in aa['byvars']:
            for exact_var, applier in adict.get('exact', {}).items():
                if var == exact_var:
                    return applier
            for regex_var, applier in adict.get('regex', {}).items():
                if re.fullmatch(regex_var, var):
                    return applier
        if missing_ok:
            return self.MODES['identity'](None, array_dims=[])
        else:
            raise SubsamplingNotFoundError(f'no subsampling applier for var {var!r}.')

    def get_var_info(self, var, default=UNSET):
        '''returns the (mode, info, array_dims, None_or_snaps_mode_and_info) for var within self.

        default: UNSET or any object
            if var not found, return default if provided, else raise SubsamplingNotFoundError
        '''
        # handle the forall stuff
        forall = self.json.get('forall', {})
        forall_array_dims = forall.get('array_dims', None)
        forall_snaps = forall.get('snaps', None)
        for vdict in self.json.get('byvars', []):
            array_dims = vdict.get('array_dims', []) if forall_array_dims is None else forall_array_dims
            snapstuff = vdict.get('snaps', None) if forall_snaps is None else forall_snaps
            for exact_var, [mode, info] in vdict.get('exact', {}).items():
                if var == exact_var:
                    return (mode, info, array_dims, snapstuff)
            for regex_var, [mode, info] in vdict.get('regex', {}).items():
                if re.fullmatch(regex_var, var):
                    return (mode, info, array_dims, snapstuff)
        if default is UNSET:
            return default
        else:
            raise SubsamplingNotFoundError(f'no subsampling info for var {var!r}.')

    # # # SAVING / LOADING # # #
    @classmethod
    def load(cls, file_or_dir=None, **kw_cls):
        '''load based on subsampling_info.txt.
        
        file_or_dir: None or str
            where to load from.
            None --> load from current directory
            directory --> load from 'subsampling_info.txt' here,
                            or from 'subsampling_info/subsampling_info.txt' here.
            file --> load from this file.
        '''
        paths = cls.subsampling_path_manager_cls.implied(file_or_dir)
        paths.assert_isfile('info_txt')
        with open(paths['info_txt'], 'r') as f:
            s = json.load(f)
        if paths.isfile('info_meta'):
            with open(paths['info_meta'], 'r') as f:
                meta = json.load(f)
        else:
            meta = None
        return cls(s, meta=meta, **kw_cls)

    def save(self, dir=os.curdir):
        '''saves subsampling info. Return SubsamplingInfoPathManager instance with relevant paths.
        Crashes if this would overwrite any existing files.

        dir: str
            save to 'subsampling_info/subsampling_info.txt' within this directory,
            or to 'subsampling_info.txt' within this dir if dir basename == 'subsampling_info'.

        Also saves 'subsampling_meta.txt' if relevant, at same level as subsampling_info.txt.
        '''
        paths = self.subsampling_path_manager_cls.implied(dir=dir, make=True)
        paths.assert_does_not_exist('info_txt')
        if self.meta is not None:
            paths.assert_does_not_exist('info_meta')
        with open(paths['info_txt'], 'w') as f:
            json.dump(self.json, f, indent=self._json_indent)
        if self.meta is not None:
            with open(paths['info_meta'], 'w') as f:
                json.dump(self.meta, f, indent=self._json_indent)
        return paths

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}({self.json!r})'

    # # # MODES -- REGISTER # # #
    MODES = {}  # dict of {mode name: applier class} for registered subsampling applier modes

    @staticmethod
    def register_applier_mode(cls_applier_to_register, mode=None):
        '''register cls_applier_to_register as a known applier mode for subsampling.'''
        cls = cls_applier_to_register
        if mode is None: mode = cls.__name__
        if mode in SubsamplingInfo.MODES:
            raise KeyError(f'mode {mode!r} already registered.')
        cls.mode = mode
        SubsamplingInfo.MODES[mode] = cls


''' --------------------- SubsamplingPathManager --------------------- '''

class SubsamplingInfoPathManager(dict):
    '''class that helps manage subsampling info paths.
    info_dir: str
        path to subsampling_info directory. Internally will be converted to abspath.

    All paths will be stored internally as abspaths.
        info_dir: subsampling_info
        info_txt: subsampling_info/subsampling_info.txt
        info_meta: subsampling_info/subsampling_meta.txt
    subclasses might add other paths too.
    Note: self.basename(key) aliases to os.path.basename(self[key]).

    make: whether to create the 'subsampling_info' folder if it doesn't exist yet.
    exist_ok: if False and any folders to make already exist, crash instead.
    '''
    basename_info_dir = 'subsampling_info'
    basename_info_txt = 'subsampling_info.txt'
    basename_info_meta = 'subsampling_meta.txt'
    keys_to_make_paths = ('info_dir',)

    def __init__(self, info_dir, *, make=True, exist_ok=True):
        if not os.path.basename(info_dir) == self.basename_info_dir:
            raise InputError(f'expected basename(info_dir)={self.basename_info_dir!r}. Got info_dir={info_dir!r}')
        super().__init__(info_dir=os.path.abspath(info_dir))
        self.init_path_strings()
        self.make = make
        self.exist_ok = exist_ok
        if make:
            self.make_paths()

    def init_path_strings(self):
        '''initialize all path strings. See help(self) for details.'''
        self['info_txt'] = os.path.join(self['info_dir'], self.basename_info_txt)
        self['info_meta'] = os.path.join(self['info_dir'], self.basename_info_meta)

    def make_paths(self):
        '''create paths for all keys in self.keys_to_make_paths, if the paths don't exist already.
        if any paths existed already but not self.exist_ok, crash before making any of them!
        '''
        if not self.exist_ok:
            for key in self.keys_to_make_paths:
                if os.path.exists(self[key]):
                    raise FileExistsError(f'key {key!r} path already exists: {self[key]!r}')
        for key in self.keys_to_make_paths:
            os.makedirs(self[key], exist_ok=self.exist_ok)

    @classmethod
    def implied(cls, file_or_dir=None, *, dir=None, make=None, **kw):
        '''returns SubsamplingInfoPathManager implied from path information or current directory.
        file_or_dir: None or str
            None --> look within current directory (or use dir if dir is provided)
            directory --> info_dir = here if 'subsampling_info.txt' contained within it,
                            else look for 'subsampling_info' directory to use as info_dir.
            file --> if 'subsampling_info.txt', use info_dir = directory containing file.
        dir: None or str
            'subsampling_info' --> info_dir = dir
            any other str --> info_dir = dir/subsampling_info
            Cannot provide dir AND file_or_dir.
        make: None or bool
            None --> use True if dir is provided, else False.

        (use cls.basename_info_txt and cls.basename_info_dir instead of 'subsampling_info.txt' and 'subsampling_info'.)
        '''
        if (file_or_dir is not None) and (dir is not None):
            raise InputConflictError('cannot provide both file_or_dir and dir.')
        kw.update(make=(dir is not None) if make is None else make)
        # provided dir (probably saving a new subsampling info)
        if dir is not None:
            if os.path.basename(dir) == cls.basename_info_dir:
                info_dir = os.path.abspath(dir)
            else:
                info_dir = os.path.join(os.path.abspath(dir), cls.basename_info_dir)
            return cls(info_dir, **kw)
        # did not provide dir (probably loading an existing subsampling info)
        if file_or_dir is None:
            file_or_dir = os.getcwd()
        if os.path.isdir(file_or_dir):
            d = os.path.abspath(file_or_dir)
            if os.path.exists(os.path.join(d, cls.basename_info_txt)):
                info_dir = d
                return cls(info_dir, **kw)
            elif os.path.exists(os.path.join(d, cls.basename_info_dir)):
                info_dir = os.path.join(d, cls.basename_info_dir)
                return cls(info_dir, **kw)
            else:
                raise FileNotFoundError(f'no {cls.basename_info_txt!r} or {cls.basename_info_dir!r} in {d!r}')
        elif os.path.isfile(file_or_dir):
            f = os.path.abspath(file_or_dir)
            if os.path.basename(f) == cls.basename_info_txt:
                info_dir = os.path.dirname(f)
                return cls(info_dir, **kw)
            else:
                raise InputError(f'expected basename(file)={cls.basename_info_txt!r} but got file={f!r}')
        else:
            raise FileNotFoundError(f'file_or_dir={file_or_dir!r} not found.')

    # # # ALIASES TO OS OPERATIONS # # #
    def basename(self, key):
        '''returns os.path.basename(self[key])'''
        return os.path.basename(self[key])

    def exists(self, key):
        '''returns os.path.exists(self[key])'''
        return os.path.exists(self[key])

    def isdir(self, key):
        '''returns os.path.isdir(self[key])'''
        return os.path.isdir(self[key])

    def isfile(self, key):
        '''returns os.path.isfile(self[key])'''
        return os.path.isfile(self[key])

    def assert_does_not_exist(self, key):
        if self.exists(key):
            raise FileExistsError(f'{self[key]!r}')
    def assert_exists(self, key):
        if not self.exists(key):
            raise FileNotFoundError(f'{self[key]!r}')
    def assert_isdir(self, key):
        if not self.isdir(key):
            raise NotADirectoryError(f'{self[key]!r}')
    def assert_isfile(self, key):
        if not self.isfile(key):
            raise FileNotFoundError(f'{self[key]!r}')

    # # # DISPLAY # # #
    def __repr__(self):
        contents = ',\n'.join(f'{k!r:>25s} : {v!r}' for k, v in self.items())
        return f'{type(self).__name__}({{\n{contents}}})'


class SubsamplingResultPathManager(SubsamplingInfoPathManager):
    '''class that helps manage paths related to results of subsampling.
    "Result of subsampling" appears in folder structure as:
        RUNDIR/subsampling_result/SNAPSDIR
        RUNDIR/subsampling_result/subsampling_info
    Contrast this with interpreting subsampled data, which expects folder structure:
        RUNDIR/SNAPSDIR (of subampled data)
        RUNDIR/subsampling_info
    The idea is that the contents of subsampling_result may be treated as a new RUNDIR.
    
    snaps_dir: str
        path to snaps directory. Internally will be converted to abspath.

    All paths will be stored internally as abspaths:
        run_dir: RUNDIR   # abspath to any directory.
        snaps_dir: RUNDIR/SNAPSDIR
        result_dir: RUNDIR/subsampling_result   # [TODO] allow for other names here as well, e.g. subsampling_result_N.
        subsampled_snaps_dir: RUNDIR/subsampling_result/SNAPSDIR
        info_dir: RUNDIR/subsampling_result/subsampling_info
        info_txt: RUNDIR/subsampling_result/subsampling_info/subsampling_info.txt
        info_meta: RUNDIR/subsampling_result/subsampling_info/subsampling_meta.txt
    subclasses might add other paths too.
    Note: self.basename(key) aliases to os.path.basename(self[key]).
    
    make: whether to create subsampling-related folders if they doesn't exist yet.
    exist_ok: if False and any folders to make already exist, crash instead.
    '''
    basename_result_dir = 'subsampling_result'
    keys_to_make_paths = ('result_dir', 'subsampled_snaps_dir', 'info_dir')

    def __init__(self, snaps_dir, make=True, exist_ok=True):
        super(SubsamplingInfoPathManager, self).__init__(snaps_dir=os.path.abspath(snaps_dir))
        self.init_path_strings()
        self.make = make
        self.exist_ok = exist_ok
        if make:
            self.make_paths()

    def init_path_strings(self):
        '''initialize all path strings. See help(self) for details.'''
        self['run_dir'] = os.path.dirname(self['snaps_dir'])
        basename_result_dir = self.basename_result_dir
        self['result_dir'] = os.path.join(self['run_dir'], basename_result_dir)
        self['subsampled_snaps_dir'] = os.path.join(self['result_dir'], self.basename('snaps_dir'))
        self['info_dir'] = os.path.join(self['result_dir'], self.basename_info_dir)
        super().init_path_strings()  # initialize info_txt and info_meta

    @classmethod
    def implied(cls, file_or_dir=None, *, dir=None, make=None, **kw):
        raise NotImplementedError(f'{type(self).__name__}.implied()')

    # # # STORAGE SIZE INFO # # #
    def nbytes_snaps_dirs(self):
        '''return dict of storage size (in bytes) of snaps_dir and subsampled_snaps_dir.
        Keys will be: 'nbytes_snaps_dir', 'nbytes_subsampled_snaps_dir'.
        Vals will be nbytes (int) if path exists, else None.
        '''
        result = {}
        for key in ('snaps_dir', 'subsampled_snaps_dir'):
            if self.exists(key):
                result[f'nbytes_{key}'] = nbytes_path(self[key])
            else:
                result[f'nbytes_{key}'] = None
        return result


# tell SubsamplingInfo about SubsamplingInfoPathManager
SubsamplingInfo.subsampling_path_manager_cls = SubsamplingInfoPathManager


''' --------------------- SubsamplingApplier --------------------- '''

class SubsamplingApplier():
    '''applier of a subsampling method.

    info: any object
        info required for self.apply.
    array_dims: list of str
        names of expected array dimensions in order.
        May be used to translate between information in info, and an array being subsampled.
    extra_info: any objects
        any extra kwargs will all be saved in self.extra_info.
    '''
    def __init_subclass__(cls, *, mode=None, **kw):
        '''register subclass as a subsampling mode.'''
        super().__init_subclass__(**kw)
        SubsamplingInfo.register_applier_mode(cls, mode=mode)

    def __init__(self, info, *, array_dims, snap_applier=None, **extra_info):
        self.info = info
        self.array_dims = array_dims
        self.snap_applier = snap_applier
        self.extra_info = extra_info
        self._check_info()

    # # # METHODS -- SUBCLASS SHOULD IMPLEMENT # # #
    def _check_info(self):
        '''checks that self.info is in the expected format. Crash if not.'''
        raise NotImplementedError(f'{type(self).__name__}._check_info()')

    def apply(self, array):
        '''apply subsampling to array; return new array.'''
        raise NotImplementedError(f'{type(self).__name__}.apply()')

    def apply1d(self, dict_):
        '''return result of subsampling dict_ of 1d arrays, with keys in array_dims.
        Subclass should raise SubsamplingError if this is not possible.
        '''
        raise NotImplementedError(f'{type(self).__name__}.apply1d()')

    # # # METHODS DEFINED HERE # # #
    def apply_snap_applier(self, snap_srcs, *, missing_ok=True):
        '''return self.snap_applier(snap_srcs), or snap_srcs unchanged if self.snap_applier=None.

        missing_ok: bool
            whether it is okay for snap_applier to be None
            False --> crash with SubsamplingNotFoundError if snap_applier is None.
        '''
        result = snap_srcs
        if self.snap_applier is None:
            if not missing_ok:
                raise SubsamplingNotFoundError('no snap_applier provided, and missing_ok=False.')
        else:
            result = self.snap_applier.apply(snap_srcs)
        return result

    def __repr__(self):
        return f'{type(self).__name__}({self.info!r}, array_dims={self.array_dims!r}, mode={self.mode!r})'


class SubsamplingIdentity(SubsamplingApplier, mode='identity'):
    '''subsampling applier that does nothing!'''
    def _check_info(self):
        '''checks that self.info is in the expected format. Crash if not.'''
        if self.info is not None:
            raise SubsamplingFormatError('info for identity should be None.')

    def apply(self, array):
        '''return array, unchanged.'''
        return array

    def apply1d(self, dict_):
        '''return dict_, unchanged.'''
        return dict_


class SubsamplingSlice(SubsamplingApplier, mode='slice'):
    '''slice subsampling applier. Expect info to be dict of {array_dim: [start, stop, step]}.'''
    # # # REQUIRED BY PARENT # # #
    def _check_info(self):
        '''checks that self.info is in the expected format. Crash if not.'''
        for key, val in self.info.items():
            if len(val) != 3:
                raise SubsamplingFormatError(f'info for {key!r} should have 3 elements: [start, stop, step].')
            if key not in self.array_dims:
                raise InputError(f'key {key!r} not in array_dims={self.array_dims!r}.')

    def apply(self, array):
        '''apply subsampling to array; return new array.'''
        array = np.asanyarray(array)
        slices = tuple(slice(*self.info.get(dim, (None,))) for dim in self.array_dims)
        return array[slices]

    def apply1d(self, dict_):
        '''return result of subsampling dict_ of 1d arrays, with keys in array_dims'''
        slices = {k: slice(*self.info.get(k, (None,))) for k in dict_}
        result = {k: v[slices[k]] for k, v in dict_.items()}
        return result

    # # # ADDITIONAL METHODS # # #
    def slice(self, dim=None):
        '''returns slice for this dim, or dict of {dim: slice} for all dims in array_dims.'''
        if dim is None:
            return {dim: slice(*self.info.get(dim, (None,))) for dim in self.array_dims}
        else:
            return slice(*self.info.get(dim, (None,)))

    def start(self, dim=None):
        '''returns slice start for this dim, or dict of {dim: slice start} for all array_dims.'''
        if dim is None:
            return {d: s.start for d, s in self.slices().items()}
        else:
            return slice(*self.info.get(dim, (None,))).start

    def stop(self, dim=None):
        '''returns slice stop for this dim, or dict of {dim: slice stop} for all array_dims.'''
        if dim is None:
            return {d: s.stop for d, s in self.slices().items()}
        else:
            return slice(*self.info.get(dim, (None,))).stop

    def step(self, dim=None):
        '''returns slice step for this dim, or dict of {dim: slice step} for all array_dims.'''
        if dim is None:
            return {d: s.step for d, s in self.slices().items()}
        else:
            return slice(*self.info.get(dim, (None,))).step


''' --------------------- Subsampler --------------------- '''

class Subsampler():
    '''interface to help with actually doing the subsampling, of a Subsamplable object.

    target: Subsamplable
        the Subsamplable object which self will subsample when called.
    info: SubsamplingInfo, str, or json-like dict
        str --> path to load subsampling info from.
                can be filename or directory. see SubsamplingInfo.load for details.
        dict --> subsampling info in json-like format, see SubsamplingInfo for details.

    Below, "snap src" refers to the hashable object that identifies a snapshot.
    Its exact behavior is not defined here and may vary across different Subsamplable subclasses.
    E.g. it may be a filepath, or a Snap object.
    The only expectation here is that each src must be hashable.

    Expects Subsamplable target object to have:
        - rawvars_loadable(src): list of all directly loadable vars within snap src.
        - rawvar_load(var, src): load a single raw var from snap src.
            OR override rawvar_load_and_subsample instead.
        - rawvar_save(var, data, dst): save a single raw var to snap dst.
        - target.snapdir: should tell abspath of folder with all snapshots.
        - target.snaps.file_path(self): should tell array of snapshot filepaths.
        - target.subsampling_info_cls: SubsamplingInfo subclass to use for this target.

    Results will be saved to subsampling result paths; see SubsamplingResultPathManager for help.
    '''
    path_manager_cls = SubsamplingResultPathManager

    def __init__(self, target, info):
        self.target = target
        self.info = target.subsampling_info_cls.implied_from(info)

    def subsample(self, *, subsampled_data_exist_ok=False):
        '''apply subsampling to all data as appropriate (determined by self.info)
        returns self.path_manager() instance with relevant paths.
            (results will be saved to paths determined by SubsamplingResultPathManager;
            probably 'subsampling_result' at the same directory-level as target.snapdir.)

        Never overwrites any of the pre-subsampling data.
        By default, refuses to overwrite any existing data at all.

        subsampled_data_exist_ok: bool
            whether it is okay for subsampled data file(s) to already exist before saving any data. Default False.
            (Doesn't interact with any subsampling_info files.)
        '''
        srcs, src2vars = self.srcs_and_src2vars()
        paths = self.path_manager(make=False)  # make=False --> don't create any folders here yet.
        updater = ProgressUpdater(wait=False)  # wait=False --> always show progress.
        for i, src in enumerate(srcs):
            updater.print(f'Subsampling snap {i+1} of {len(srcs)}: {src!r}')
            self.subsample_snap(src, include_vars=src2vars.get(src, None),
                                dst_exists_ok=subsampled_data_exist_ok)
        self.info.update_meta(paths.nbytes_snaps_dirs())  # put info about snaps dirs sizes
        self.info.save(paths['info_dir'])
        updater.finalize('subsample()')
        return paths

    def path_manager(self, *, make=True, exist_ok=True):
        '''returns instance of path_manager_cls, based on self.target.snapdir.
        Use this when creating subsampled data.

        make: whether to create the 'subsampling_result' folder if it doesn't exist yet.
        exist_ok: if False and any folders to make already exist, crash instead.
        '''
        snapdir = self.target.snapdir  # in its own line to make debugging easier in case of crash.
        return self.path_manager_cls(snapdir, make=make, exist_ok=exist_ok)

    # # # GETTING SRCS AND VARS FOR EACH SRC # # #
    def srcs(self):
        '''returns list of all snap srcs in self.target (before applying any subsampling)'''
        return self.target.subsampling_snap_srcs()

    def srcs_and_src2vars(self):
        '''returns (srcs, src2vars), where:
            srcs = [list of all snap srcs to keep]
            src2vars = dict of {src: [vars to keep for src]} if relevant.
                        if keeping all vars from src, src2vars might exclude src key.
                        (if keeping all vars from all srcs, src2vars may be an empty dict.)
        '''
        if self.info.has_forall_snaps():
            srcs = self.srcs_forall()  # returns list of srcs
            src2vars = {}  # no var-specific mapping based on snaps
        elif self.info.has_byvars_snaps():
            var2srcs = self.srcs_byvars()   # returns dict of {var: [srcs to keep for var]}
            srcs = set.union(*[set(v) for v in var2srcs.values()])
            src2vars = {src: set() for src in srcs}
            for var, srcs_ in var2srcs.items():
                for src in srcs_:
                    src2vars[src].add(var)
        else:  # NO SNAP SUBSAMPLING
            srcs = self.srcs()  # returns list of srcs
            src2vars = {}  # no var-specific mapping based on snaps
        return srcs, src2vars

    def srcs_forall(self):
        '''list of all srcs, subsampled by self.info['forall']['snaps'].'''
        if not self.info.has_forall_snaps():
            raise SubsamplingError('srcs_forall expects info to have "forall" snaps')
        srcs = self.srcs()
        return self.info.get_applier('snaps').apply(srcs)

    def srcs_byvars(self):
        '''dict of {var: [srcs in which to keep data for var]}

        Subsampling will be relative to full list of snaps;
            raise SubsamplingError if would need to write var at a snap where it was previously missing.
            E.g. crash if var5 appears only every 5 snaps but info says slice to every 7 snaps for var5.

        Any var with no snap subsampling indicated will be kept in all srcs where it was already.
        '''
        if not self.info.has_byvars_snaps():
            raise SubsamplingError('srcs_byvars expects info to have "byvars" snaps')
        srcs = self.srcs()
        src2vars = self.target.rawvars_loadable_across(srcs)
        var2srcs = {}  # {var: [srcs that have var]}
        for s, vars_ in src2vars.items():
            for v in vars_:
                var2srcs.setdefault(v, set()).add(s)
        # getting result: {var: [srcs to keep for var]}
        result = {}
        SKIPPED = set()
        for var in var2srcs:
            try:  # any subsampling for this var? if no, just keep it in all srcs where it was already.
                applier = self.info.get_applier(var, missing_ok=False)
            except SubsamplingNotFoundError:
                result[var] = var2srcs[var]
                SKIPPED.add(var)  # bookkeeping to help with debugging
                continue
            try:  # any snap subsampling for this var? if no, just keep it in all srcs where it was already.
                subsampled = applier.apply_snap_applier(srcs, missing_ok=False)
            except SubsamplingNotFoundError:
                result[var] = var2srcs[var]
                SKIPPED.add(var)  # bookkeeping to help with debugging
                continue
            # crash if subsampling would cause needing to write a var at a snap that previously didn't have it.
            for s in subsampled:
                if var not in src2vars[s]:
                    raise SubsamplingError(f'var {var!r} does not appear in original snap {s!r}')
            # actually subsampling this var! Add to result:
            result[var] = subsampled
        return result

    # # # SUBSAMPLING SINGLE SNAP # # #
    def subsample_snap(self, src, *, dst=None, dst_exists_ok=False, include_vars=None):
        '''apply subsampling to all appropriate data vars in this snap.
        returns abspath to dst where data was saved (or, None if no data was saved).

        src: any hashable object
            snap src to subsample.
        dst: None or str
            filepath for where to save subsampled data.
            None --> self.dst_from_src(src)
        include_vars: None or list of strs
            list of vars to include in result.
            None --> use self.target.rawvars_loadable(src)
                    (and do not apply any 'snaps' subsampling from subsampling_info)
            if empty list, do not save any vars, and return None instead of path.
        '''
        # setup / bookkeeping
        if include_vars is None:
            include_vars = self.target.rawvars_loadable(src)
        elif len(include_vars) == 0:
            return None
        if dst is None:
            dst = self.dst_from_src(src)
        if (not dst_exists_ok) and os.path.exists(dst):
            raise FileExistsError(f'{dst!r} already exists and dst_exists_ok=False.')
        # do the subsampling
        for var in include_vars:
            subsampled = self.target.rawvar_load_and_subsample(var, src, self.info)
            self.target.rawvar_save(var, subsampled, dst)
        return os.path.abspath(dst)

    def dst_from_src(self, src):
        '''return filepath for where to save subsampled data for this src.
        Details determined by self.path_manager() and self.target.snap_src_to_filepath().
        '''
        paths = self.path_manager()
        src_path = self.target.snap_src_to_filepath(src)
        dst = os.path.join(paths['subsampled_snaps_dir'], os.path.basename(src_path))
        return dst

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}(target={self.target!r},\n\ninfo={self.info!r})'


''' --------------------- Subsamplable --------------------- '''

class Subsamplable():
    '''object that can be subsampled.
    
    Subclass should implement:
        - rawvars_loadable(src): list of all directly loadable vars within snap src.
        - rawvar_load(var, src): load a single raw var from snap src.
            OR override rawvar_load_and_subsample(var, src, info) to avoid using rawvar_load.
        - rawvar_save(var, data, dst): save a single raw var to snap dst.
            Before overwriting any existing data, subclass's implementation should crash by default.
        - _snap_src_to_filepath(self, src): return filepath for snap src.
        - self.snapdir: should tell abspath of folder with all snapshots.
            OR, override subsampling_path_manager() to avoid using snapdir.
        - self.snaps.file_path(self): should tell array of snapshot filepaths.
            OR, override subsampling_snap_srcs to avoid using self.snaps.file_path(self).
    Subclass may use str for snap src but might use something else.
    The only requirement for snap src imposed here is that it must be hashable.
    '''
    subsampler_cls = Subsampler
    subsampling_info_cls = SubsamplingInfo
    subsampling_info_path_manager_cls = SubsamplingInfoPathManager

    # # # METHODS -- REQUIRED OR SUGGESTED, FOR SUBCLASS TO OVERRIDE # # #
    def subsampling_snap_srcs(self):
        '''list of all snap srcs in self (before applying any subsampling)'''
        return self.snaps.file_path(self)

    def snap_src_to_filepath(self, src):
        '''return filepath for snap src. See also: self.subsampling_snap_srcs().
        Subclass must implement this.
        '''
        raise NotImplementedError(f'{type(self).__name__}.snap_src_to_filepath()')

    def rawvars_loadable(src):
        '''returns list of all directly loadable vars within snap src.
        Subclass must implement this.
        '''
        raise NotImplementedError(f'{type(self).__name__}.rawvars_loadable()')

    def rawvars_loadable_across(self, srcs=None):
        '''returns dict of {src: [vars loadable from src]} for all snap srcs
        The implementation here just calls rawvars_loadable for each src.

        if srcs is None, use self.subsampling_snap_srcs().
        '''
        if srcs is None: srcs = self.subsampling_snap_srcs()
        result = {}
        for src in srcs:
            loadable = self.rawvars_loadable(src)
            result[src] = loadable
        return result

    def rawvar_load(var, src):
        '''load a single raw var from snap src.
        Subclass can implement this or override rawvar_load_and_subsample instead.
        '''
        raise NotImplementedError(f'{type(self).__name__}.rawvar_load()')

    def rawvar_save(var, data, dst):
        '''save a single raw var to snap dst.
        Subclass must implement this.
        Before overwriting any existing data, subclass's implementation should crash by default.
        '''
        raise NotImplementedError(f'{type(self).__name__}.rawvar_save()')

    def rawvar_load_and_subsample(self, var, src, info):
        '''returns var loaded and subsampled from src.
        Subclass should either:
            - implement rawvar_load(var, src)
            - override rawvar_load_and_subsample instead,
                e.g. slice h5py object instead of loading full array to slice.
        '''
        applier = info.get_applier(var)
        data = self.rawvar_load(var, src)
        return applier.apply(data)

    # # # SUBSAMPLING (to help with subsampling some data) # # #
    def subsampler(self, subsampling_info):
        '''return self.subsampler_cls instance with self and subsampling_info.
        See help(self.subsample) for details on subsampling_info.
        '''
        return self.subsampler_cls(self, subsampling_info)

    def subsample(self, subsampling_info, *, subsampled_data_exist_ok=False):
        '''apply subsampling to all data indicated by subsampling_result_info.
        returns SubsamplingResultPathManager with relevant paths.
            (results will be saved to paths determined by SubsamplingResultPathManager;
            probably 'subsampling_result' at the same directory-level as target.snapdir.)

        Never overwrites any of the pre-subsampling data.
        By default, refuses to overwrite any existing data at all.
        
        subsampling_info: SubsamplingInfo, str, or json-like dict
            str --> path to load subsampling info from.
                    can be filename or directory. see SubsamplingInfo.load for details.
            dict --> subsampling info in json-like format, see SubsamplingInfo for details.
        subsampled_data_exist_ok: bool
            whether it is okay for subsampled data file(s) to already exist before saving any data. Default False.
            (Doesn't interact with any subsampling_info files.)
        '''
        subsampler = self.subsampler(subsampling_info)
        paths = subsampler.subsample(subsampled_data_exist_ok=subsampled_data_exist_ok)
        return paths

    # # # SUBSAMPLING INFO (to help with reading subsampled data) # # #
    # [TODO] reload subsampling_info when relevant, but cache results to know when it's not relevant.
    subsampling_info = simple_property('_subsampling_info', setdefaultvia='_default_subsampling_info',
            doc='''SubsamplingInfo telling how existing data was subsampled from prior data.
            By default, this info will be loaded from {self.dirname}/subsampling_info.''')

    @subsampling_info.setter
    def subsampling_info(self, info):
        '''set self._subsampling_info to info.
        see self.subsampling_info_cls.implied_from(info) for more details.
        '''
        self._subsampling_info = self.subsampling_info_cls.implied_from(info)
    def _default_subsampling_info(self):
        '''default value for self.subsampling_info: load from {self.dirname}/subsampling_info.'''
        try:
            return self.load_subsampling_info()
        except FileNotFoundError:
            return None

    def load_subsampling_info(self):
        '''load subsampling info from {self.dirname}/subsampling_info.'''
        return self.subsampling_info_cls.load(self.dirname)



"""
File Purpose: quantities known when quasineutral=True
"""

from .quantity_loader import QuantityLoader

class QuasineutralLoader(QuantityLoader):
    '''quantities that require quasineutral=True.
    Usually, this just provides a different way to get an electron quantity,
        e.g. electron number density using assumption of quasineutrality.
    '''
    _quasineutral = False
    _assert_QN = True   # whether to assert self.quasineutral before doing any operation here.

    def __init__(self, quasineutral=None, **kw_super):
        if quasineutral is not None: self.quasineutral = quasineutral
        super().__init__(**kw_super)

    cls_behavior_attrs.register('quasineutral', default=False)
    @property
    def quasineutral(self):
        '''tells whether self is in quasineutral mode.'''
        return self._quasineutral
    @quasineutral.setter
    def quasineutral(self, value):
        '''set self.quasineutral, then call self.on_changed_quasineutral(old=old_value, new=new_value)'''
        old_qn = self._quasineutral
        if value is not old_qn:
            self._quasineutral = value
            self.on_changed_quasineutral(old=old_qn, new=value)

    def on_changed_quasineutral(self, *, old, new):
        '''called when self.quasineutral changes.
        default behavior: do nothing.
        '''
        pass

    def assert_QN(self):
        '''asserts self.quasineutral. if not self._assert_QN, instead does not assert self.quasineutral.'''
        if self._assert_QN and not self.quasineutral:
            raise AssertionError(f'Expected {type(self).__name__}.quasineutral=True, got {self.quasineutral}.')

    @known_var(deps=['nq', 'q'], ignores_dims=['fluid'])
    def get_ne(self):
        '''electron number density. (ne qe) + sum_i (ni qi) = 0. (using qe < 0 convention)'''
        self.assert_QN()
        niqi = self('nq', fluid=self.fluids.ions())
        qe = self('q', fluid=self.fluids.get_electron())
        return niqi / -qe



"""
File Purpose: multifluid analysis from single-fluid mhd.
"""
from .mhd_fluids import (
    ElementHandler, MhdFluid, ElementHaver,
    Specie, IonMixture,
    ElementHandlerList, MhdFluidList, ElementHaverList,
    SpecieList, IonMixtureList,
    SPECIES, ION_MIXTURES, ION_MIXTURE_SPECIES,
)
from .mhd_genrad_tables import GenradTable, GenradTableManager
from .mhd_multifluid_bases import MhdMultifluidBasesLoader
from .mhd_multifluid_calculator import MhdMultifluidCalculator
from .mhd_multifluid_densities import MhdMultifluidDensityLoader
from .mhd_multifluid_ionization import MhdMultifluidIonizationLoader, saha_n1n0



"""
File Purpose: tools related to fft (of numpy or xarray arrays).

This subpackage might depend on other modules in tools.
"""
from .array_fft import (
    fftN, fft2, fft1,
    fftfreq_shifted,
    ifftN,
    ifftfreq_shifted,
)
from .fft_dimnames import FFTDimname
from .fft_slices import FFTSlices
from .xarray_fft import (
    xarray_fftN,
    xarray_ifftN,
    xarray_lowpass,
)



"""
File Purpose: misc tests for eppic hookup in PlasmaCalcs.
These tests require some simulation output.
Not intended to read lots of different values,
    just to read a few specific/misc. values to check if they work properly.
"""
import os
import pytest

# setting mp start method to fork (instead of 'spawn') prevents leaked semaphore objects. Not sure why.
# it also makes the tests a bit faster, maybe...
import multiprocessing as mp
mp.set_start_method('fork', force=True)

import numpy as np
import xarray as xr

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)


''' --------------------- can create eppic calculator --------------------- '''

def get_eppic_calculator(**kw_init):
    '''make an eppic calculator'''
    with pc.InDir(os.path.join(HERE, 'test_eppic_tinybox')):
        ec = pc.EppicCalculator.from_here(**kw_init)
    return ec

def test_make_eppic_calculator():
    '''ensure can make an eppic calculator (with various inputs to __init__).'''
    get_eppic_calculator()
    get_eppic_calculator(snaps_from='timers')
    get_eppic_calculator(kw_units=dict(M=1))

''' --------------------- can load moments --------------------- '''

def test_load_moment():
    '''ensure we can load moment related values'''
    ec = get_eppic_calculator()  # defined earlier in this file
    ec('moment1')               # just checking to ensure these don't crash
    ec('moment2')               
    ec('moment3')               
    ec('moment4')

''' --------------------- runtimes --------------------- '''

def test_reading_runtimes():
    '''test reading runtimes from eppic timers.dat'''
    ec = get_eppic_calculator()
    with pytest.raises(ValueError):
        ec('runtimes')  # fails because snap 0 file was deleted but timer is known for it.
    ec = get_eppic_calculator(snaps_from='timers')
    arr = ec('runtimes')  # just checking that this doesn't crash!
    arr = ec('calc_timestep_cost')
    assert arr.size == len(ec.snap)

def test_reading_clocktimes():
    '''test reading clock_times from logfile.'''
    ec = get_eppic_calculator(snaps_from='timers')
    # ensure can read clock times
    ec('clock_times')
    # this run took 19 seconds
    # (according to start & stop times. Total job time was actually 23 seconds.
    #   but 4 seconds of that might have been slurm processing stuff.)
    assert ec('total_seconds', item=True) == 19
    # timers.dat units are (experimentally found to be) roughly 1 [timer unit] == 0.01 [seconds].
    assert abs(100 - (ec('run_time').sum() / ec('steps_seconds'))) < 5


''' --------------------- multi_slices --------------------- '''

def test_reading_multi_slices():
    '''test reading when using multi_slices.
    Also confirms Dataset.pc.size works properly.
    '''
    ec = get_eppic_calculator()
    with ec.using(multi_slices=dict(ndim=1, ikeep=0)):
        ds = ec('n')
    with ec.using(slices=dict(x=0)):
        arr_x0 = ec('n')
    with ec.using(slices=dict(y=0)):
        arr_y0 = ec('n')
    assert np.all(ds['keep_x'] == arr_y0)  # keep x <--> used y=0 slice
    assert np.all(ds['keep_y'] == arr_x0)  # keep y <--> used x=0 slice
    assert ds.pc.size == arr_x0.pc.size + arr_y0.pc.size


''' --------------------- enable_fromfile=False --------------------- '''

def test_enable_fromfile():
    '''test reading when using enable_fromfile=False.'''
    ec = get_eppic_calculator()
    # ensure can load normally
    m0 = ec('m', enable_fromfile=True)
    n0 = ec('n', enable_fromfile=True)
    # when load_direct_fromfile disabled, can load m (doesn't depend on file) but not n.
    m1 = ec('m', enable_fromfile=False)
    assert m1.identical(m0)
    with pytest.raises(pc.QuantCalcError):
        _should_crash = ec('n', enable_fromfile=False)
    # double-check that setting attribute via calculator produces the same behavior
    ec.enable_fromfile = True
    m2 = ec('m')
    assert m2.identical(m0)
    n2 = ec('n')
    assert n2.identical(n0)
    ec.enable_fromfile = False
    m3 = ec('m')
    assert m3.identical(m0)
    with pytest.raises(pc.QuantCalcError):
        _should_crash2 = ec('n')


''' --------------------- multiprocessing + parenthesis --------------------- '''

def test_mp_with_parenthesis():
    '''ensure can run multiprocessing with parenthesis.'''
    # it can be tricky because parenthesis_memory is a class variable,
    # and pickling doesn't include changes to class or global variables, by default.
    # this ensures that the __getstate__ & __setstate__ solution works properly
    ec = get_eppic_calculator()
    ec.ncpu = 2
    ec('(n0)')
    ec('(deltafrac_n+1)*7')
    ec('mean_(n)')
    ec('nmean_(u**2)', stats_dimpoint_wise=False)
    ec('nmean_(u**2)', stats_dimpoint_wise=True)

## [TODO] DISABLED TEST_MP_TIMEOUT UNTIL FURTHER NOTICE
## I'm tired of gitlab runner failing this test inconsistently;
## it seems like sometimes gitlab CI can handle it but other times it can't.
## last time it crashed, the logs looked like:
# tests/test_eppic/test_eppic_reading_misc.py::test_mp_with_parenthesis PASSED
# tests/test_eppic/test_eppic_reading_misc.py::test_mp_timeout [test_mp_timeout] started
# [test_mp_timeout] got eppic calculator
# [test_mp_timeout] about to check ec(longvar, ncpu=1) for 10 snaps
# [test_mp_timeout] ec(longvar, ncpu=1) for 10 snaps took 1.46e-01 seconds
# [test_mp_timeout] setting NSNAP=684 to make computations take "a long time" (approx 10 seconds).
# [test_mp_timeout] starting ec(longvar, ncpu=1, timeout=1) for 684 snaps
# Terminated
# WARNING: step_script could not run to completion because the timeout was exceeded.
## I gave step_script 5 mins total to run, so 10 seconds should have completed.
## Most likely the timeout code itself is causing the error;
## maybe sometimes gitlab doesn't like SIGALRM??
## Can re-enable if you want to debug. Probably do it in a separate branch though.

# takes roughly 2 seconds, but run after test_mp_with_parenthesis,
#   so that we don't run this until we know ncpu>1 actually works on this machine.
# def test_mp_timeout():
#     '''ensure eppic calculator will timeout properly.'''
#     # this test has print statements because pytest will print them if the test fails,
#     #   and this test seems to fail inconsistently on some of the gitlab runners.
#     #   the prints will hopefully help debug why it fails on some runners but not others.
#     print(f'[test_mp_timeout] started', flush=True)  # flush=True --> print ASAP.
#     ec = get_eppic_calculator()
#     print(f'[test_mp_timeout] got eppic calculator', flush=True)
#     longvar = 'stats_lowpass_(u_drift**2)'  # something that takes a while to calculate
#     # check time it takes to calculate:
#     ec.snap = slice(10)
#     assert len(ec.snap) == 10  # there are at least 10 snaps in eppic_tinybox.
#     print(f'[test_mp_timeout] about to check ec(longvar, ncpu=1) for 10 snaps', flush=True)
#     with pc.Stopwatch(printing=False) as watch:
#         ec(longvar, ncpu=1, snap=ec.snap[:10])
#     time10 = watch.time_elapsed()
#     print(f'[test_mp_timeout] ec(longvar, ncpu=1) for 10 snaps took {time10:.2e} seconds', flush=True)
#     time_per_snap = time10 / 10
#     safety = 10
#     timeout = 1
#     NSNAP = int(safety * timeout / time_per_snap)  # should take roughly [safety * timeout] seconds.
#     print(f'[test_mp_timeout] setting NSNAP={NSNAP} to make computations take "a long time" (approx 10 seconds).', flush=True)
#     ec.snap = [i%10 for i in range(NSNAP)]
#     assert time10 * NSNAP/10 > 1.1  # ensure longvar actually takes > 1s to calculate, when doing NSNAP snaps.
#     # check timeout when ncpu=1 (i.e., single processing)
#     with pc.Stopwatch(printing=False) as watch:
#         print(f'[test_mp_timeout] starting ec(longvar, ncpu=1, timeout=1) for {NSNAP} snaps', flush=True)
#         with pytest.raises(pc.TimeoutError):
#             ec(longvar, ncpu=1, timeout=1)   # ensure timeout works when ncpu=1
#     print(f'[test_mp_timeout] ec(longvar, ncpu=1, timeout=1) for {NSNAP} snaps, timeout after {watch.time_elapsed():.2e} seconds', flush=True)
#     assert len(ec.snap) == NSNAP  # ensure original snaps restored even after TimeoutError.
#     # check timeout when ncpu=2 (i.e., multiprocessing)
#     with pc.Stopwatch(printing=False) as watch:
#         print(f'[test_mp_timeout] starting ec(longvar, ncpu=2, timeout=1) for {NSNAP} snaps', flush=True)
#         with pytest.raises(pc.TimeoutError):
#             ec(longvar, ncpu=2, timeout=1)   # ensure timeout works when ncpu>1
#     print(f'[test_mp_timeout] ec(longvar, ncpu=2, timeout=1) for {NSNAP} snaps, timeout after {watch.time_elapsed():.2e} seconds', flush=True)


''' --------------------- misc small multiprocessing --------------------- '''
# e.g., ensure https://gitlab.com/Sevans7/plasmacalcs/-/issues/2 does not appear again;
# that issue was: Can't pickle local object 'EppicBasesLoader._get_T_neutral.<locals>.Tn_loader'

def test_mp_small_misc():
    '''ensure small misc vars do not crash when ncpu > 1.'''
    ec = get_eppic_calculator()
    ec.ncpu = 2
    ec('T_n')
    ec('moment1')
    ec('moment2')
    ec('moment3')
    ec('moment4')
    # ec('vdist', fluid=0)  # would be nice to check, but vdist is not saved in tinybox.



"""
File purpose: plot limits, e.g. vlims, xlim, ylim.
"""
import matplotlib.pyplot as plt
import numpy as np

from ..plot_settings import PlotSettings
from ...errors import InputConflictError, InputMissingError, InputError
from ...tools import (
    finite_min, finite_max, finite_percentile,
)

_paramdocs_lims = {
    'skipna': '''bool
        whether to return None instead of nan if min or max would be nan.''',
    'plot_settings': '''None or PlotSettings instance
        if margin is None, get from plot_settings (or matplotlib defaults if None in plot_settings)''',
    'data_interval': '''None or [min, max]
        if provided, use this instead of the actual range of data from the plot.''',
    'scale': '''None, 'linear', or 'log'
        whether margin is in linear or log space. If None, use ax's current scale.''',
}

''' --------------------- COMPUTE LIMS --------------------- '''

@PlotSettings.format_docstring()
def calculate_vlims(values, *, vmin=None, vmax=None, robust=False, to_dict=False, expand_flat=True):
    '''return the calculated vmin, vmax, for all of these values.

    values: array-like or list of array-like objects
        the values to use for calculating the vlims.
        if list of array-like objects, the array-like objects may be of different shapes.
    vmin, vmax: None or value
        if provided, use this value for vmin, vmax. Otherwise, calculate it.
    robust: {robust}
    to_dict: bool
        if True, return dict(vmin=vmin, vmax=vmax) instead of (vmin, vmax)

    expand_flat: bool
        if True, and computed vmin > vmax, and provided vmin or vmax but not both,
        then adjust the not-provided value such that vmin > vmax. I.e.:
            provided vmin --> use vmax = vmin + 0.05.  (this matches matplotlib default when vmin=vmax)
            provided vmax --> use vmin = vmax - 0.05.  (this matches matplotlib default when vmin=vmax)
    '''
    if robust is True:
        robust = PlotSettings.get_default('robust')  # get robust percentile.
    provided = dict(vmin=vmin, vmax=vmax)
    if vmin is None:
        if robust:
            if getattr(values, 'dtype', None) == bool:
                vmin = 0  # booleans have no percentiles; vmin=0 always.
            else:
                vmin = finite_percentile(values, robust)
        else:
            vmin = finite_min(values)
    if vmax is None:
        if robust:
            if getattr(values, 'dtype', None) == bool:
                vmax = 1  # booleans have no percentiles; vmax=1 always.
            else:
                vmax = finite_percentile(values, 100 - robust)
        else:
            vmax = finite_max(values)
    if expand_flat and vmin > vmax:
        if provided['vmin'] is not None:
            vmax = vmin + 0.05
        elif provided['vmax'] is not None:
            vmin = vmax - 0.05
    return dict(vmin=vmin, vmax=vmax) if to_dict else (vmin, vmax)

@PlotSettings.format_docstring(sub_ntab=1)
def update_vlims(values, d, *, in_place=False):
    '''calculate vmin, vmax of values, and return copy of d updated appropriately.
    values: array-like or list of array-like objects
        the values to use for calculating the vlims.
        if list of array-like objects, the array-like objects may be of different shapes.
    d: dict which may have keys 'vmin', 'vmax', and/or 'robust', indicating:
        vmin, vmax: None or value
            if provided, use this value for vmin, vmax. Otherwise, calculate it.
        robust: {robust}
        
        Note: the 'robust' key will be popped from the result.
    in_place: bool, default False
        whether to perform the operation in place. True --> edit d directly.
    '''
    if not in_place:
        d = d.copy()
    robust = d.pop('robust', False)
    vmin = d.pop('vmin', None)
    vmax = d.pop('vmax', None)
    vmin, vmax = calculate_vlims(values, vmin=vmin, vmax=vmax, robust=robust)
    d['vmin'] = vmin
    d['vmax'] = vmax
    return d

@PlotSettings.format_docstring(**_paramdocs_lims)
def calculate_lims_from_margin(min, max, margin=None, *, x=None, skipna=False, scale=None):
    '''calculate the lims to use, based on min, max, and margin.
    Return (min, max) after scaling them by this margin.

    min, max: None or number
        None --> return (None, None). (in this case, provided min and max must both be None.)
        number --> treat this as the before-applying-margin limits (in data-coordinates).
    margin: {margin}
    x: None, 'x', or 'y'
        must be provided if `margin` is not; tells which matplotlib default to use.
        ignored if `margin` is provided.
    skipna: {skipna}
    scale: {scale}
        (Actually, if None then default to 'linear' scale, in this function.)
    '''
    if min is None and max is None:
        return (None, None)
    elif min is None or max is None:
        errmsg = f'expected min and max to both be None or both be non-None, but got min={min!r}, max={max!r}.'
        raise InputConflictError(errmsg)
    # else, min and max are both non-None
    # bookkeeping
    if margin is None:
        margin = _get_margin_default(x=x)
    if scale is None:
        scale = 'linear'
    if scale not in ('linear', 'log'):
        raise InputError(f'expected scale to be None, "linear", or "log", but got scale={scale!r}.')
    if scale == 'log':
        min = np.log10(min)
        max = np.log10(max)
    # >> apply margin <<
    range_ = max - min
    max_result = max + range_ * margin
    min_result = min - range_ * margin
    # bookkeeping
    if skipna:
        if np.isnan(min_result): min_result = None
        if np.isnan(max_result): max_result = None
    if scale == 'log':
        min_result = 10 ** min_result
        max_result = 10 ** max_result
    return (min_result, max_result)

def _get_margin_default(*, margin=None, x=None, plot_settings=None):
    '''return the margin to use, based on margin and plot_settings.
    margin: None or number
        the margin to use. returned immediately if not None.
        if necessary, get it from matplotlib defaults, or from plot_settings if possible.
        raise InputError if failed to get margin.
    x: None, 'x', or 'y'
        get '{x}margin' from plot_settings. ('margin' if x is None)
        Or, from matplotlib defaults.
    plot_settings: None or PlotSettings
        get margin from plot_settings if provided, and margin is None.
        if plot_settings.get(f'{x}margin' is None, also try getting plot_settings.get('margin').
            if that is also None, get from matplotlib defaults.
    '''
    if margin is not None:
        return margin
    if plot_settings is not None:
        if x is None:
            margin = plot_settings.get('margin')
            if margin is None:
                raise InputMissingError('must provide "x", "margin", or non-None "margin" in plot_settings.')
        else:
            margin = plot_settings.get(f'{x}margin')
            if margin is None:
                margin = plot_settings.get('margin')
        if margin is not None:
            return margin
    # if reached this line, get from matplotlib defaults.
    if x is None:
        raise InputMissingError('must provide x or margin.')
    if x != 'x' and x != 'y':
        raise InputError(f'expected x to be None, "x", or "y", but got x={x!r}.')
    return plt.rcParams[f'axes.{x}margin']

def get_data_interval(x, ax=None):
    '''return the minimum interval containing all data on the plot axes.
    Returned as [min, max] regardless of axis orientation.

    x: 'x' or 'y'
        str telling the axis to get the data interval for.
    ax: None or matplotlib axis object
        the Axes to get the data interval from. None --> ax=plt.gca().
        will use ax.xaxis or ax.yaxis, depending on `x`.
    '''
    if x == 'x':
        axis = 'xaxis'
    elif x == 'y':
        axis = 'yaxis'
    else:
        raise InputError(f'expected x to be "x" or "y", but got x={x!r}.')
    if ax is None:
        ax = plt.gca()
    axis = getattr(ax, axis)
    return axis.get_data_interval()

def _get_scale_default(*, scale, x, ax=None):
    '''returns the scale to use, based on scale, x, and ax.'''
    if scale is not None:
        return scale
    if ax is None:
        ax = plt.gca()
    if x == 'x':
        scale = ax.get_xscale()
    elif x == 'y':
        scale = ax.get_yscale()
    else:
        raise InputError(f'expected x to be "x" or "y", but got x={x!r}.')
    return scale

@PlotSettings.format_docstring(**_paramdocs_lims)
def get_lims_with_margin(margin=None, *, x, ax=None, plot_settings=None,
                         data_interval=None, skipna=False, scale=None):
    '''returns lims to use for axis x, based on margin and the current data interval.

    margin: {margin}
    x: 'x' or 'y'
        str telling the axis to get the data interval for.
    ax: None or matplotlib axis object
        the Axes to get the data interval from. None --> ax=plt.gca().
        will use ax.xaxis or ax.yaxis, depending on `x`.
    plot_settings: {plot_settings}
    data_interval: {data_interval}
    skipna: {skipna}
    scale: {scale}
    '''
    if data_interval is None:
        data_min, data_max = get_data_interval(x, ax=ax)
    else:
        data_min, data_max = data_interval
    margin = _get_margin_default(margin=margin, x=x, plot_settings=plot_settings)
    scale = _get_scale_default(scale=scale, x=x, ax=ax)
    return calculate_lims_from_margin(data_min, data_max, margin=margin, skipna=skipna, scale=scale)


''' --------------------- SET LIMS --------------------- '''

@PlotSettings.format_docstring(**_paramdocs_lims)
def plt_zoom(x='both', margin=None, *, ax=None, plot_settings=None,
             data_interval=None, skipna=True, scale=None):
    '''zoom in/out to view the data with the specified margin.
    Default is to zoom both axes such that all data is viewable, with a small margin.

    x: 'both', 'x', or 'y'
        axis for which to set the margin.
        if None, set margin for xaxis and yaxis.
    margin: {margin}
    ax: None or matplotlib axis object
        the Axes on which to set the margin(s). None --> ax=plt.gca().
    plot_settings: {plot_settings}
    data_interval: {data_interval}
    skipna: {skipna}
    scale: {scale}

    returns the new plt.xlim() or plt.ylim().
    If x is None, returns (plt.xlim(), plt.ylim()) instead.
    '''
    _x_input = x  # save original x to help with debugging.
    if ax is None:
        ax = plt.gca()
    xvals = ['x', 'y'] if x=='both' else [x]
    result = []
    for x in xvals:
        lims = get_lims_with_margin(margin=margin, x=x, ax=ax, plot_settings=plot_settings,
                                    data_interval=data_interval, skipna=skipna, scale=scale)
        if x == 'x':
            result.append(ax.set_xlim(lims))
        elif x == 'y':
            result.append(ax.set_ylim(lims))
        else:
            raise InputError(f'expected x to be None, "x", or "y", but got x={x!r}.')
    return result[0] if len(xvals)==1 else result

@PlotSettings.format_docstring(**_paramdocs_lims)
def plt_zoomx(margin=None, *, ax=None, plot_settings=None,
              data_interval=None, skipna=True, scale=None, **kw_set_margin):
    '''zoom in/out along the x axis. Equivalent: plt_zoom(x='x', ...)

    margin: {margin}
    ax: None or matplotlib axis object
        the Axes on which to set the margin. None --> ax=plt.gca().
    plot_settings: {plot_settings}
    data_interval: {data_interval}
    skipna: {skipna}
    scale: {scale}

    returns the new plt.xlim()
    '''
    return plt_zoom(x='x', margin=margin, ax=ax, plot_settings=plot_settings,
                    data_interval=data_interval, skipna=skipna, scale=scale)

@PlotSettings.format_docstring(**_paramdocs_lims)
def plt_zoomy(margin=None, *, ax=None, plot_settings=None,
              data_interval=None, skipna=True, scale=None, **kw_set_margin):
    '''zoom in/out along the y axis. Equivalent: plt_zoom(x='y', ...)

    margin: {margin}
    ax: None or matplotlib axis object
        the Axes on which to set the margin. None --> ax=plt.gca().
    plot_settings: {plot_settings}
    data_interval: {data_interval}
    skipna: {skipna}
    scale: {scale}

    returns the new plt.ylim()
    '''
    return plt_zoom(x='y', margin=margin, ax=ax, plot_settings=plot_settings,
                    data_interval=data_interval, skipna=skipna, scale=scale)


''' --------------------- SET LIMS (FROM PLOT SETTINGS) --------------------- '''

def set_margins_if_provided(plot_settings, *, ax=None):
    '''sets the margins for a plot, if they are non-None in plot_settings.
    see also: plt_set_margin.

    ax: None or matplotlib axis object
        the Axes on which to set the margin. None --> ax=plt.gca().
    '''
    margin = plot_settings.get('margin', default=None)
    for x in ('x', 'y'):
        xmargin = plot_settings.get(f'{x}margin', default=margin)
        if xmargin is not None:
            plt_zoom(x=x, margin=xmargin, ax=ax)

def set_lims_if_provided(plot_settings, *, ax=None):
    '''sets plt.xlim and/or plt.ylim, if they are non-None in plot_settings.

    ax: None or matplotlib axis object
        the Axes on which to set the lims. None --> ax=plt.gca().

    returns (bool, bool) indicating whether xlim, ylim were set.
    '''
    xlim = plot_settings.get('xlim', default=None)
    ylim = plot_settings.get('ylim', default=None)
    set_x = xlim is not None
    set_y = ylim is not None
    if set_x: ax.set_xlim(xlim)
    if set_y: ax.set_ylim(ylim)
    return (set_x, set_y)



"""
File Purpose: tools for easily managing attributes
"""

import functools

from ..properties import weakref_property_simple, alias
from ..sentinels import ATTR_UNSET, UNSET
from ...defaults import DEFAULTS


''' --------------------- very simple context managers --------------------- '''

class CallInContext():
    '''context manager which calls f with *args and **kw upon entry, and does nothing upon exit.'''
    def __init__(self, f, *args, **kw):
        self.f = f
        self.args = args
        self.kw = kw

    def __enter__(self):
        self.f(*self.args, **self.kw)

    def __exit__(self, exc_type, exc_value, traceback):
        pass


''' --------------------- attrs context managers --------------------- '''

# # # HELPERS # # #
def set_or_del_attr(obj, attr, value, unset=ATTR_UNSET, *, unset_fail_ok=True):
    '''usually, setattr(obj, attr, value). However, if `value` is `unset`, delattr(obj, attr) instead.
    (`value` will be compared to `unset` using "is", not "==".)

    unset_fail_ok: bool
        whether to ignore AttributeError when `value` is `unset` but obj.attr is not defined.
    '''
    if value is unset:
        try:
            delattr(obj, attr)
        except AttributeError:
            if not unset_fail_ok:
                raise
    else:
        setattr(obj, attr, value)

# # # SIMPLE CONTEXT MANAGERS & ALIASES # # #
def maintaining_attrs(self, *attrs, **attrs_as_flags):
    '''returns context manager which restores attrs of self to their original values, upon exit.
    E.g. maintaining_attrs(obj, 'attr1', 'attr2', attr3=True, attr4=False)
    --> will restore upon exit, original values of obj.attr1, attr2, and attr3, but not attr4.
    '''
    return MaintainingAttrs(self, *attrs, **attrs_as_flags)

def using_attrs(self, attrs_as_dict=dict(), _unset_sentinel=ATTR_UNSET, **attrs_and_values):
    '''returns context manager which sets attrs of obj upon entry; restores original values upon exit.
    _unset_sentinel: any value, default ATTR_UNSET
        upon entry, delete any attrs with value _unset_sentinel (compared via 'is').
        E.g. using_attrs(obj, _unset_sentinel=None, x=None) --> del obj.x upon entry.
    '''
    return UsingAttrs(self, attrs_as_dict, _unset_sentinel=_unset_sentinel, **attrs_and_values)

def maintain_attrs(*attrs):
    '''return decorator which restores attrs of obj after running function.
    It is assumed that obj is the first arg of function.
    '''
    def attr_restorer(f):
        @functools.wraps(f)
        def f_but_maintain_attrs(obj, *args, **kwargs):
            '''f but attrs are maintained.'''
            __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
            with MaintainingAttrs(obj, *attrs):
                return f(obj, *args, **kwargs)
        return f_but_maintain_attrs
    return attr_restorer

def use_attrs(attrs_as_dict=dict(), **attrs_and_values):
    '''return decorator which sets attrs of object before running function then restores them after.
    It is assumed that obj is the first arg of function.
    '''
    def attr_setter_then_restorer(f):
        @functools.wraps(f)
        def f_but_set_then_restore_attrs(obj, *args, **kwargs):
            '''f but attrs are set beforehand then restored afterward.'''
            __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
            with UsingAttrs(obj, attrs_as_dict, **attrs_and_values):
                return f(obj, *args, **kwargs)
        return f_but_set_then_restore_attrs
    return attr_setter_then_restorer

class MaintainingAttrs():
    '''context manager which restores attrs of obj to their original values, upon exit.
    This includes deleting any of those attrs upon exit which did not exist upon entry.

    kwargs can be provided with key=bool; attr key will be included if True.
        E.g. MaintainingAttrs(obj, 'attr1', 'attr2', attr3=True, attr4=False)
        --> will maintain obj.attr1, obj.attr2, and obj.attr3, but not obj.attr4.
    '''
    _unset_sentinel = ATTR_UNSET   # the sentinel used to indicate an attribute was not set.

    def __init__(self, obj, *attrs, **attrs_as_flags):
        self.obj = obj
        self.attrs = list(attrs)
        self.attrs.extend(attr for attr, flag in attrs_as_flags.items() if flag)

    def obj_current_attrs(self):
        '''return dict of {attr: obj.attr} for current values of attrs.'''
        return {attr: getattr(self.obj, attr, self._unset_sentinel) for attr in self.attrs}

    def enter_handle_attrs(self):
        '''take all steps related to attrs which should occur upon entry.
        Provided separately from self.__enter__ for access to low-level functionality.
        '''
        self.memory = self.obj_current_attrs()
        self._inside_attrs = True

    def exit_handle_attrs(self):
        '''take all steps related to attrs which should occur upon exit.
        Provided separately from self.__exit__ for access to low-level functionality.
        '''
        if DEFAULTS.DEBUG > 10:
            if len(self.memory)>0: print('   > restoring attrs:', self.memory)
        for attr, val in self.memory.items():
            set_or_del_attr(self.obj, attr, val, unset=self._unset_sentinel)
        self._inside_attrs = False

    def inside_attrs(self):
        '''tells whether currently inside attrs handling for self. None if never entered; False if exited.'''
        return getattr(self, '_inside_attrs', None)

    def __enter__(self):
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.enter_handle_attrs()

    def __exit__(self, exc_type, exc_value, traceback):
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.exit_handle_attrs()

    # # # DISPLAY # # #
    def __repr__(self):
        contents = f'managing attrs {[attr for attr in self.attrs]}; inside_attrs={self.inside_attrs()}'
        return f'{type(self).__name__}({contents})'


class UsingAttrs(MaintainingAttrs):
    '''context manager which sets attrs of obj upon entry; restores original values upon exit.
    This includes deleting any of those attrs upon exit which did not exist upon entry.

    _unset_sentinel: any value, default ATTR_UNSET
        upon entry, delete any attrs with value _unset_sentinel (compared via 'is').
        E.g. UsingAttrs(obj, _unset_sentinel=None, x=None) --> del obj.x upon entry.
    '''
    def __init__(self, obj, attrs_as_dict=dict(), *, _unset_sentinel=ATTR_UNSET, **attrs_and_values):
        self.obj = obj
        self.attrs = {**attrs_as_dict, **attrs_and_values}
        self._unset_sentinel = _unset_sentinel
        if DEFAULTS.DEBUG > 10:
            if len(self.attrs)>0: print('   < setting attrs:', self.attrs)

    def enter_handle_attrs(self):
        '''take all steps related to attrs which should occur upon entry.
        Provided separately from self.__enter__ for access to low-level functionality.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        super().enter_handle_attrs()
        for attr, val in self.attrs.items():
            set_or_del_attr(self.obj, attr, val, unset=self._unset_sentinel)

    # __enter__ and __exit__ inherited from super().


# # # LESS-SIMPLE CONTEXT MANAGERS # # #
class UsingAttrsSignaled(UsingAttrs):
    '''context manager which (enter) sets attrs of obj; (exit) restores original values; (both) sends signals (e.g. to obj).

    signals are sent at entry and exit, using whichever of the following are provided:
        signal_enter: None or str
            immediately BEFORE setting attrs, obj.signal_enter(current, new) during ENTRY
        signal_exit: None or str
            immediately BEFORE setting attrs, obj.signal_exit(current, new) during EXIT
        signal_context_shift: None or str
            immediately BEFORE setting attrs, obj.signal_context_shift(current, new) during ENTRY and EXIT.

        signal_entered: None or str
            immediately AFTER setting attrs, obj.signal_entered(current, new) during ENTRY
        signal_exited: None or str
            immediately AFTER setting attrs, obj.signal_exited(current, new) during EXIT
        signal_context_shifted: None or str
            immediately AFTER setting attrs, obj.signal_context_shifted(current, new) during ENTRY and EXIT.

        where (current, new) are dicts for the attrs (or a single value if len(self.attrs)==1; see `squeeze`):
            current = dict of {attr: obj.attr} for values before setting attrs.
            new = dict of {attr: obj.attr} for values after setting attrs.

    _unset_sentinel: any value, default ATTR_UNSET
        used to indiate an attribute does not exist.
        values in the dicts passed to signals will be _unset_sentinel if the attribute does not exist.
        upon entry, delete any attrs with value _unset_sentinel (compared via 'is').
        E.g. UsingAttrsSignaled(obj, _unset_sentinel=None, x=None) --> del obj.x upon entry.
    squeeze: bool, default True
        if True, and len(self.attrs)==1, then current and new will be single values, not dicts.
    signal_target: None or object
        if provided, use this object instead of self.obj, for sending signals.

    --- EXAMPLES ---
        # starting with obj.depth==0:
        with UsingAttrsSignaled(obj.depth=1, signal_context_shifted='_on_set_depth'):
            # during entry, obj._on_set_depth(0, 1), immediately after setting obj.depth=1.
            # during exit, obj._on_set_depth(depth_pre_exit, 0), immediately after setting obj.depth=0,
            #   where depth_pre_exit = obj.depth just before exiting.

        # starting with obj.depth==0:
        with UsingAttrsSignaled(obj.depth=1, signal_context_enter='_on_increment_depth', squeeze=False):
            # during entry, obj._on_increment_depth({depth: 0}, {depth: 1}), immediately before setting obj.depth=1.

        # starting with obj.depth==0, obj.x==3
        with UsingAttrsSignaled(obj.depth=1, x=7, signal_context_exit='_on_exit_depth_and_x'):
            # during exit, first obj._on_exit_depth_and_x({depth: obj.depth, x: obj.x}, {depth: 0, x: 3}),
            #   immediately before setting obj.depth=0 and obj.x=3
    '''
    def __init__(self, obj, attrs_as_dict=dict(),
                 signal_enter=None, signal_exit=None, signal_context_shift=None,
                 signal_entered=None, signal_exited=None, signal_context_shifted=None,
                 _unset_sentinel=ATTR_UNSET, squeeze=True, signal_target=None,
                  **attrs_and_values):
        super().__init__(obj, attrs_as_dict, _unset_sentinel=_unset_sentinel, **attrs_and_values)
        self.signal_enter = signal_enter
        self.signal_exit = signal_exit
        self.signal_context_shift = signal_context_shift
        self.signal_entered = signal_entered
        self.signal_exited = signal_exited
        self.signal_context_shifted = signal_context_shifted
        self.squeeze = squeeze
        self.signal_target = signal_target

    def obj_maybe_signal(self, name, current, new):
        '''possibly send signal to obj, based on mode and signal_* attributes.
        name: str, one of ('enter', 'exit', 'context_shift', 'entered', 'exited', 'context_shifted')
        current: dict
            dict of {attr: obj.attr} for values before setting attrs due to shifting context.
        new: dict
            dict of {attr: obj.attr} for values after setting attrs due to shifting context.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        signal = getattr(self, f'signal_{name}')
        if signal is not None:
            target = self.obj if self.signal_target is None else self.signal_target
            current = next(iter(current.values())) if self.squeeze and len(current)==1 else current
            new     = next(iter(new.values()))     if self.squeeze and len(new)==1     else new
            fsignal = getattr(target, signal)
            return fsignal(current, new)

    def __enter__(self):
        current = self.obj_current_attrs()
        new = self.attrs.copy()
        self.obj_maybe_signal('enter', current, new)
        self.obj_maybe_signal('context_shift', current, new)
        super().__enter__()
        self.obj_maybe_signal('entered', current, new)
        self.obj_maybe_signal('context_shifted', current, new)
    
    def __exit__(self, exc_type, exc_value, traceback):
        current = self.obj_current_attrs()
        new = self.memory.copy()
        self.obj_maybe_signal('exit', current, new)
        self.obj_maybe_signal('context_shift', current, new)
        super().__exit__(exc_type, exc_value, traceback)
        self.obj_maybe_signal('exited', current, new)
        self.obj_maybe_signal('context_shifted', current, new)

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return list of contents to go in repr of self.'''
        contents = []
        contents.append(f'attrs.keys()={list(self.attrs.keys())}')
        for key in ('enter', 'exit', 'context_shift', 'entered', 'exited', 'context_shifted'):
            if getattr(self, f'signal_{key}') is not None:
                contents.append(f'signal_{key}={getattr(self, f"signal_{key}")!r}')
        contents.append(f'_unset_sentinel={self._unset_sentinel!r}')
        contents.append(f'squeeze={self.squeeze!r}')
        if self.signal_target is not None:
            contents.append(f'signal_target={self.signal_target!r}')

    def __repr__(self):
        return f'{type(self).__name__}({", ".join(self._repr_contents())})'


''' --------------------- other attr management --------------------- '''

class IncrementableAttrManager():
    '''an incrementable attribute, and methods to manage other attributes when it is incremented or decremented.'''
    _unset_sentinel = ATTR_UNSET   # the sentinel used to indicate an attribute was not set.

    def __init__(self, obj, default=0, *, step=1):
        self.obj = obj
        self.default = default
        self.step = step
        if step <= 0:
            raise NotImplementedError(f'nonpositive step, inside IncrementableAttrManager. Got step={step}.')

        self.value = default

        # {value: list of (context, remove_after_decrementing_to) pairs}
        #   use all contexts in list when incrementing or decrementing self.value to value.
        #   when decrementing to remove_after_decrementing_to, remove the context from the list.
        self._using_at = dict()

    obj = weakref_property_simple('_obj')  # behaves like obj but is weakref, to avoid circular references.

    incrementable_attr_manager_value = alias('value')   # useful for more verbose debugging

    def __int__(self):
        '''return int(self.value)'''
        return int(self.value)

    def clear(self):
        '''set self.value to self.default, and empty self._using_at.'''
        self.value = self.default
        self._using_at.clear()

    def increment(self):
        '''context manager which increments self.value upon entry, and restores old value upon exit.
        CAUTION: this is a context manager; it will not increment self.value until entered,
            via "with" statement, e.g. "with self.increment(): ...".
        '''
        old = self.value
        new = old + self.step
        return UsingAttrsSignaled(self,
                  incrementable_attr_manager_value=new,  # equivalent to value=new, but more verbose in debugging.
                  signal_entered='_on_incremented',
                  signal_exited='_on_decremented')

    def use_obj_attrs_at(self, value, *, remove_after_decrementing_to=UNSET, **attrs_and_values):
        '''tell self to set attrs of obj when self.value == value.
        More precisely:
            - when self incremented OR decremented to value,
                UsingAttrs(self.obj, attrs_and_values).enter_handle_attrs()
            - when self incremented beyond value OR decremented beyond value,
                UsingAttrs(...).exit_handle_attrs()

        remove_after_decrementing_to: UNSET, None, or int
            - when self decremented to remove_after_decrementing_to,
                also remove these "use_obj_attrs_at" commands.
            Note: this should always be < value, otherwise will raise ValueError.
            UNSET --> use self.default.
            None --> never remove these commands as a result of decrementing.
            int --> remove these commands as soon as value is decremented to this value.

        Note: to decrease chances of mistakes / ambiguity,
            raise ValueError if value <= self.value, or if value <= remove_after_decrementing_to
            value <= self.value, OR
            value <= remove_after_decrementing_to.

        (This method is only responsible for putting these attrs and values into self._using_at;
            the "real work" happens in self.incremented, _on_incremented, and _on_decremented.)
        '''
        rmv = remove_after_decrementing_to
        rmv = self.default if rmv is UNSET else rmv
        if (value <= self.value): #or ((rmv is not None) and (value < rmv)):
            errmsg = (f'Cannot tell self to use obj attrs at value, when value <= self.value.'
                        f' Got value={value}, self.value={self.value}.')
            raise ValueError(errmsg)
        elif (rmv is not None) and (value <= rmv):
            errmsg = (f'Cannot tell self to use obj attrs at value, when value <= remove_after_decrementing_to.'
                        f' Got value={value}, remove_after_decrementing_to={rmv}.')
            raise ValueError(errmsg)
        context = UsingAttrs(self.obj, attrs_and_values)
        self._using_at.setdefault(value, []).append((context, rmv))

    def _on_incremented(self, prev, new):
        '''called when self.value is incremented, i.e. when entering the 'incremented' context.'''
        for context, _rmv in self._using_at.get(prev, []):
            context.exit_handle_attrs()
        for context, _rmv in self._using_at.get(new, []):
            context.enter_handle_attrs()

    def _on_decremented(self, prev, new):
        '''called when self.value is decremented, i.e. when exiting the 'incremented' context.'''
        for context, _rmv in self._using_at.get(prev, []):
            context.exit_handle_attrs()
        for context, _rmv in self._using_at.get(new, []):
            context.enter_handle_attrs()

        self._remove_commands_after_decrementing_to(new)

    def _remove_commands_after_decrementing_to(self, value):
        '''remove commands from self.using_at after decrementing self to value.
        Removes any (context, rmv) pair for which rmv==value.
        If any of these contexts have not been exited, raise AssertionError.
        '''
        for _value, commands in tuple(self._using_at.items()):
            to_remove = [i for i, (context, rmv) in enumerate(commands) if rmv == value]
            for i in to_remove:  # crash with error message if any to_remove contexts have not been exited.
                context, rmv = commands[i]
                if context.inside_attrs():
                    errmsg = (f'context {context} has not been exited, but is being removed from self._using_at.'
                              '\nCrashing, then clearing self._using_at to avoid further errors.')
                    self._using_at.clear()
                    raise AssertionError(errmsg)
            if len(to_remove) == len(commands):  # remove all commands for this value
                del self._using_at[_value]
            else:  # only remove some commands for this value.
                for i in reversed(to_remove):
                    del commands[i]

    def use_obj_attrs_at_next(self, **attrs_and_values):
        '''tell self to set attrs of obj when self.increment() is next entered.
        Equivalent to self.use_obj_attrs_at(self.value + self.step, **attrs_and_values),
            with remove_after_decrementing_to = self.value.
        '''
        self.use_obj_attrs_at(self.value + self.step, remove_after_decrementing_to=self.value, **attrs_and_values)

    # # # CONTEXT MANAGERS # # #
    def using_obj_attrs_at(self, value, **attrs_and_values):
        '''context manager which tells self to set attrs of obj when self.value == value.
        Equivalent to calling self.use_obj_attrs_at(...) upon entry (and doing nothing upon exit).
        '''
        return CallInContext(self.use_obj_attrs_at, value, **attrs_and_values)

    def using_obj_attrs_at_next(self, **attrs_and_values):
        '''context manager which tells self to set attrs of obj when self.increment() is next entered.
        Equivalent to calling self.use_obj_attrs_at_next(...) upon entry (and doing nothing upon exit).
        '''
        return CallInContext(self.use_obj_attrs_at_next, **attrs_and_values)

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}(value={self.value})'

    # # # PICKLING # # #
    def __getstate__(self):
        '''return state for pickling. (pickle can't handle weakrefs. pickling is required by multiprocessing.)'''
        state = self.__dict__.copy()
        # follow weakrefs. state['_obj'] might be a weakref;
        #   self.obj would be the result of following this weakref (see weakref_property_simple).
        state['_obj'] = self.obj
        return state

    def __setstate__(self, state):
        '''set state from pickling. (pickle can't handle weakrefs. pickling is required by multiprocessing.)'''
        self.__dict__.update(state)
        # set up weakrefs. state['_obj'] might be a obj but should be a weakref instead;
        #   setting self.obj = state['_obj'] internally stores the value as a weakref.
        self.obj = state['_obj']


"""
File Purpose: loading single-fluid Muram quantities related to Equation Of State (EOS).
"""
import os

from ...errors import SnapValueError
from ...mhd import MhdEosLoader


''' --------------------- MuramEosLoader--------------------- '''

class MuramEosLoader(MhdEosLoader):
    '''single-fluid Bifrost quantities related to Equation of State (EOS): ne, T, P.

    The implementation here assumes tables available at table=self.tabin[var],
        for var='ne', 'T', or 'P', and each having a table.interp(r=r, e=e) method,
        which gives value of var in 'raw' units.
    '''

    # non-aux functionality is inherited from MhdEosLoader.
    # 'aux' functionality is implemented here: read directly from aux files.

    EOS_MODE_OPTIONS = {**MhdEosLoader.EOS_MODE_OPTIONS,
        'aux': '''read directly from aux files for eosP, eosT, and eosne.'''
    }

    def _all_eos_aux_files_exist(self):
        '''returns whether eos aux files (eosP, eosT, and eosne) exist for all snaps in self.'''
        try:
            loadable = self.directly_loadable_vars()
        except SnapValueError:
            return False  # [TODO] this is overly restrictive...
        else:
            return all(var in loadable for var in ('eosT', 'eosP', 'eosne'))

    def _default_eos_mode(self):
        '''default for how to handle "Equation of State" related variables (ne, T, P).
        (provides default value for self.eos_mode.)

        result will be 'aux' if files for 'eosT', 'eosP', and 'eosne' exist for each snap,
        else 'table' if 'tabparams.in' file exists,
        else 'ideal'.
        '''
        if self._all_eos_aux_files_exist():
            return 'aux'
        elif os.path.isfile(os.path.join(self.dirname, 'tabparams.in')):
            return 'table'
        else:
            return 'ideal'

    # tell super().get_ne to use self('ne_aux') if eos_mode=='aux':
    _EOS_MODE_TO_NE_VAR = {**MhdEosLoader._EOS_MODE_TO_NE_VAR, 'aux': 'ne_aux'}

    # tell super().get_T to use self('T_aux') if eos_mode=='aux':
    _EOS_MODE_TO_T_VAR = {**MhdEosLoader._EOS_MODE_TO_T_VAR, 'aux': 'T_aux'}

    # tell super().get_P to use self('P_aux') if eos_mode=='aux':
    _EOS_MODE_TO_P_VAR = {**MhdEosLoader._EOS_MODE_TO_P_VAR, 'aux': 'P_aux'}

    @known_var(dims=['snap'])
    def get_ne_aux(self):
        '''electron number density, from 'eosne' file.'''
        ufactor = self.u('n', convert_from='cgs')
        return self.load_maindims_var_across_dims('eosne', u=ufactor, dims=['snap'])

    @known_var(dims=['snap'])
    def get_T_aux(self):
        '''temperature, from 'eosT' file.'''
        # note: multifluid T_aux assumes same T for all fluids.
        return self.load_maindims_var_across_dims('eosT', u='K', dims=['snap'])

    @known_var(dims=['snap'])
    def get_P_aux(self):
        '''pressure, from 'eosP' file.'''
        ufactor = self.u('pressure', convert_from='cgs')
        return self.load_maindims_var_across_dims('eosP', u=ufactor, dims=['snap'])



"""
File purpose: tools related to colors (see also: colorbar.py)

End-users: see help(PlasmaCalcs.get_cmap); you'll probably just use that function.
"""

import collections

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt


from ...tools import (
    simple_property,
    NO_VALUE,
)
from ...errors import InputConflictError


''' --------------------- BaseColormap --------------------- '''

ColormapExtremes = collections.namedtuple('ColormapExtremes', ('under', 'over', 'bad'))

class BaseColormap(mpl.colors.Colormap):
    '''matplotlib.colors.Colormap class with more methods, and some properties.'''
    def __init__(self, *args_super, **kw_super):
        super().__init__(*args_super, **kw_super)
        self._default_extremes = self.extremes

    # # # PROPERTIES / METHODS FOR HANDLING EXTREMES # # #
    colorbar_extend_settings = simple_property('_colorbar_extend_settings', setdefault=dict,
            doc='''the current dict(under=bool, over=bool) corresponding to colorbar_extend.''')

    @property
    def colorbar_extend(self):
        '''defaults instructions for whether to extend colorbars using this colormap.
        One of the following values:
            'neither' --> do not extend colorbar.
            'both' --> extend colorbar in both directions.
            'min' --> extend colorbar in the direction of the minimum value.
            'max' --> extend colorbar in the direction of the maximum value.
        When setting self.colorbar_extend=value, value can be:
            str --> use value (must be 'neither', 'both', 'min', or 'max').
            True --> use 'both'
            False --> use 'neither'
        '''
        settings = self.colorbar_extend_settings
        under = settings.get('under', False)
        over = settings.get('over', False)
        if under and over: return 'both'
        elif under: return 'min'
        elif over: return 'max'
        else: return 'neither'
    @colorbar_extend.setter
    def colorbar_extend(self, value):
        '''set the value of colorbar_extend. See help(type(self).colorbar_extend) for details.'''
        if value is True:
            value = 'both'
        elif value is False:
            value = 'neither'
        if value not in ('neither', 'both', 'min', 'max'):
            raise ValueError(f'invalid value for colorbar_extend: {value}')
        settings = self.colorbar_extend_settings
        if value == 'both':
            settings['under'], settings['over'] = True, True
        elif value == 'min':
            settings['under'], settings['over'] = True, False
        elif value == 'max':
            settings['under'], settings['over'] = False, True
        elif value == 'neither':
            settings['under'], settings['over'] = False, False

    under = property(lambda self: self.get_under(),
                     lambda self, val: self.set_under(val),
            doc='''the color for low out-of-range values.
            when setting self.under = value, value can be:
                str --> use this color.
                None --> use the default color (bottom of the colormap).
                if not None, also sets self.colorbar_extend_settings['under'],
                    to True if (str or True), else False.''')
    def set_under(self, color=None, alpha=None, **kw_super):
        '''set the color for low out-of-range values. See help(type(self).under) for details.'''
        color_bool = True if not (color is None or color is False) else color
        if any(color is b_ for b_ in (None, True, False)):
            color = self._default_extremes.under
        super().set_under(color, alpha, **kw_super)
        if isinstance(color_bool, bool):
            self.colorbar_extend_settings['under'] = color_bool

    over = property(lambda self: self.get_over(),
                    lambda self, val: self.set_over(val),
                    lambda self: self.del_over(),
            doc='''the color for high out-of-range values.
            when setting self.over = value, value can be:
                str --> use this color.
                None or bool --> use the default color (top of the colormap).
                if not None, also sets self.colorbar_extend_settings['over'],
                    to True if (str or True), else False.''')
    def set_over(self, color=None, alpha=None, **kw_super):
        '''set the color for high out-of-range values. See help(type(self).over) for details.'''
        color_bool = True if not (color is None or color is False) else color
        if any(color is b_ for b_ in (None, True, False)):
            color = self._default_extremes.over
        super().set_over(color, alpha, **kw_super)
        if isinstance(color_bool, bool):
            self.colorbar_extend_settings['over'] = color_bool

    bad = property(lambda self: self.get_bad(),
                   lambda self, val: self.set_bad(val),
                   lambda self: self.del_bad(),
            doc='''the color for masked values. Use self.bad=None to restore default.''')
    def set_bad(self, color=None, alpha=None, **kw_super):
        '''set the color for masked values. See help(type(self).bad) for details.'''
        if color is None:
            color = self._default_extremes.bad
        super().set_bad(color, alpha, **kw_super)

    extremes = property(lambda self: self.get_extremes(),
                        lambda self, value: self._set_extremes_implied(),
            doc='''(under, over, bad), as a ColormapExtremes namedtuple.
            setting self.extremes = value is possible, but to avoid ambiguity:
                value must have keys "under", "over", and/or "bad",
                or attributes "under", "over", and/or "bad".
            e.g. self.extremes = dict(under='blue')
                (equivalent to self.under='blue')
            Use None to reset to default, e.g. self.extremes = None,
                (equivalent to self.under=None, self.over=None, self.bad=None)
            or for any keys, e.g. self.extremes = dict(under='blue', over=None)
                (equivalent to self.under='blue', self.over=None, but not adjusting self.bad)''')

    def get_extremes(self):
        '''return (under, over, bad) colors, as a ColormapExtremes namedtuple.'''
        return ColormapExtremes(self.under, self.over, self.bad)

    def _set_extremes_implied(self, extremes):
        '''set under, over, bad colors, based on extremes. see help(type(self).extremes) for details.'''
        if extremes is None:
            self.set_extremes(under=None, over=None, bad=None)
            return
        has_keys = False
        result = dict()
        # check keys, first
        for under in ('under', 'over', 'bad'):
            try:
                val = extremes[under]
            except (KeyError, TypeError):
                pass
            else:
                result[under] = val
                has_keys = True
        if not has_keys:
            # check attrs (only if no keys were found)
            for under in ('under', 'over', 'bad'):
                try:
                    under = getattr(extremes, key)
                except AttributeError:
                    pass
                else:
                    result[key] = under
        self.set_extremes(**result)

    def set_extremes(self, *, under=NO_VALUE, over=NO_VALUE, bad=NO_VALUE):
        '''set under, over, bad colors, based on keywords.
        see help(type(self).extremes) for details.
        '''
        if under is not NO_VALUE:
            self.set_under(under)
        if over is not NO_VALUE:
            self.set_over(over)
        if bad is not NO_VALUE:
            self.set_bad(bad)

    def __copy__(self):
        '''return a copy of self. (used by self.copy()'''
        result = super().__copy__()
        result._default_extremes = self._default_extremes
        result.colorbar_extend_settings = self.colorbar_extend_settings.copy()
        return result

    def with_extremes(self, *, under=NO_VALUE, over=NO_VALUE, bad=NO_VALUE):
        '''return a copy of self, with new colors for under, over, and/or bad.
        see help(type(self).extremes) for details.
        '''
        result = self.copy()
        result.set_extremes(under=under, over=over, bad=bad)
        return result

    def with_under(self, under):
        '''return a copy of self, with new color for under.'''
        return self.with_extremes(under=under)
    def with_over(self, over):
        '''return a copy of self, with new color for over.'''
        return self.with_extremes(over=over)
    def with_bad(self, bad):
        '''return a copy of self, with new color for bad.'''
        return self.with_extremes(bad=bad)

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return list of contents to go in repr of self.'''
        contents = []
        if self.name is not None:
            contents.append(f'name={self.name!r}')
        if self.N is not None:
            contents.append(f'N={self.N}')
        if self.colorbar_extend != 'neither':
            contents.append(f'colorbar_extend={self.colorbar_extend!r}')
        if not np.all(self.under == self._default_extremes.under):
            contents.append(f'under={self.under!r}')
        if not np.all(self.over == self._default_extremes.over):
            contents.append(f'over={self.over!r}')
        if not np.all(self.bad == self._default_extremes.bad):
            contents.append(f'bad={self.bad!r}')
        return contents

    def __repr__(self):
        '''return repr(self)'''
        contents = ", ".join(self._repr_contents())
        return f'{type(self).__name__}({contents})'


''' --------------------- Colormap (from ListedColormap) --------------------- '''

class Colormap(BaseColormap, mpl.colors.ListedColormap):
    '''matplotlib.colors.ListedColormap with some extra methods.
    Especially helpful options include:
        self.at(fractions) --> new Colormap with only the colors at fractions
        self.shrink(min, max) --> new Colormap but smaller range of colors
        self + other --> new Colormap from concatenating lists of colors
    
    under: NO_VALUE, None, str, or other specifier for a color.
        the color for low out-of-range values.
        if provided (not NO_VALUE), sets self.under = under.
    over: NO_VALUE, None, str, or other specifier for a color.
        the color for high out-of-range values.
        if provided (not NO_VALUE), sets self.over = over.
    bad: NO_VALUE, None, str, or other specifier for a color.
        the color for masked values.
        if provided (not NO_VALUE), sets self.bad = bad.

    see help(mpl.colors.ListedColormap) for more details on __init__.
    '''
    # # # CREATION / INITIALIZATION # # #
    def __init__(self, *args_super, under=NO_VALUE, over=NO_VALUE, bad=NO_VALUE, **kw_super):
        super().__init__(*args_super, **kw_super)
        self.set_extremes(under=under, over=over, bad=bad)

    @classmethod
    def from_cmap(cls, cmap, **kw):
        '''return Colormap from cmap.
        cmap: Colormap or matplotlib.colors.ListedColormap
            [TODO] or matplotlib.colors.LinearSegmentedColormap
        additional kwargs go to cls.__new__ and cls.__init__.
        '''
        if kw.get('under', NO_VALUE) is NO_VALUE: kw['under'] = cmap.get_under()
        if kw.get('over',  NO_VALUE) is NO_VALUE: kw['over']  = cmap.get_over()
        if kw.get('bad',   NO_VALUE) is NO_VALUE: kw['bad']   = cmap.get_bad()
        try:
            colors = cmap.colors
        except AttributeError:  # LinearSegmentedColormap...
            colors = cmap(np.linspace(0, 1, cmap.N))
        return cls(colors, name=cmap.name, N=cmap.N, **kw)

    @classmethod
    def get_cmap(cls, based_on=None, N=None, *, colors=None, name=None, reversed=False,
                 under=NO_VALUE, over=NO_VALUE, bad=NO_VALUE,
                 **kw_init):
        '''Get a Colormap instance, possibly based on a default colormap from matplotlib.
        based_on: matplotlib.colors.Colormap, str, or None
            the colormap to base the result on.
            None --> use the default colormap (from matplotlib rc; image.cmap)
                    (or, if colors are provided, use colors to make a new colormap instead).
            str --> use the colormap with this name from matplotlib.
                    (or, if colors are provided, set this to the name of the new colormap).
            Colormap --> base the result on this colormap.
                    (mutually exclusive with providing a value for colors)
        N: None or int
            number of colors to use.
            None --> use the number of colors in based_on.
            int --> resample to this many colors.
        colors: None or array-like of colors
            None --> ignore this parameter
            array-like --> provide these to make a new Colormap.
                    (mutually exclusive with providing a Colormap for based_on)
        name: None or str
            the name for the new Colormap.
            if None, infer name from the old colormap or the provided value of based_on.
        reversed: bool
            whether to return result.reversed(), instead.
        under: NO_VALUE, None, str, or other specifier for a color.
            the color for low out-of-range values.
            if provided (not NO_VALUE), sets result.under = under.
        over: NO_VALUE, None, str, or other specifier for a color.
            the color for high out-of-range values.
            if provided (not NO_VALUE), sets result.over = over.
        bad: NO_VALUE, None, str, or other specifier for a color.
            the color for masked values.
            if provided (not NO_VALUE), sets result.bad = bad.
        
        additional kwargs are passed to cls(...) when creating the result.
        '''
        if based_on is None:
            base_name, base_cmap = None, None
        elif isinstance(based_on, str):
            base_name, base_cmap = based_on, None
        else:
            base_name, base_cmap = None, based_on
        if colors is not None and base_cmap is not None:
            raise InputConflictError('cannot provide both Colormap (for based_on) and colors.')
        if base_cmap is None:
            # get matplotlib colormap
            if colors is None:
                base_cmap = plt.get_cmap(base_name, N)
            else:
                base_cmap = mpl.colors.ListedColormap(colors, name=base_name, N=N)
        result = cls.from_cmap(base_cmap, under=under, over=over, bad=bad, **kw_init)
        if name is not None:
            result.name = name
        if N is not None:
            result = result.resampled(N)
        if reversed:
            result = result.reversed()
        return result

    # # # matplotlib isn't nice about subclassing, so we must redefine methods here: # # #
    def resampled(self, N, **kw__super):
        '''return a copy of self, resampled to size N.'''
        result = super().resampled(N, **kw__super)
        return self.from_cmap(result)

    def reversed(self, name=None, **kw__super):
        '''return a copy of self, with reversed colors.
        if name is None, use name = self.name + "_r".
        '''
        result = super().reversed(name=name, **kw__super)
        return self.from_cmap(result)

    def at(self, fractions):
        '''return copy of self but only keep colors at self(fractions).
        E.g. self.at([0.1, 0.5, 0.9]) gives cmap with N=3,
            and colors at 10%, 50%, and 90% of the way along self.
        '''
        colors = self(fractions)
        self_attrs = dict(name=self.name, under=self.under, over=self.over, bad=self.bad)
        return type(self)(colors, N=len(fractions), **self_attrs)

    def shrink(self, min=0, max=1, N=None):
        '''return copy of self but only keep colors between min and max.
        E.g. self.shrink(0.2, 0.7) only keeps colors between 20% and 70% of the way along self.
        N: None or int
            number of colors to use for the result.
            None --> keep similar spacing of colors as in self
        '''
        if not (0 <= min <= max <= 1):
            raise ValueError(f'expect 0 <= min <= max <= 1, but got min={min}, max={max}')
        if N is None:
            N = int(self.N * (max - min))  # [TODO] maybe +-1 to get to same exact spacing as self.
        colors = self(np.linspace(min, max, N))
        self_attrs = dict(name=self.name, under=self.under, over=self.over, bad=self.bad)
        return type(self)(colors, N=N, **self_attrs)

    def expand(self, min=0, max=1):
        '''return copy of self but larger, filling pre-min with min and post-max with max.
        E.g. self.expand(-0.3, 1.2) has N = 1.5 * self.N,
            with first 3/15 colors being min, and last 2/15 colors being max.
        '''
        if not (min <= 0 <= 1 <= max):
            raise ValueError(f'expect min <= 0 <= 1 <= max, but got min={min}, max={max}')
        if min == 0 and max == 1:
            return self
        pre_N = int(self.N * (0 - min))
        post_N = int(self.N * (max - 1))
        eval_at = np.concatenate([np.zeros(pre_N), np.linspace(0, 1, self.N), np.ones(post_N)])
        colors = self(eval_at)
        self_attrs = dict(name=self.name, under=self.under, over=self.over, bad=self.bad)
        return type(self)(colors, N=len(eval_at), **self_attrs)

    # # # combining colormaps # # #
    def __add__(self, other):
        '''combine self and other colormap in the intuitive way.
        result will have colors from self then other.
        result.under will be self.under.
        result.over will be other.over.
        result.bad will be self.bad.
        result.name will be '{self.name}+{other.name}', e.g. 'viridis+plasma'

        [TODO] it would be nice to enable an option to "fade" in the middle,
            for a smoother transition between the two colormaps...
        '''
        if not isinstance(other, Colormap):
            errmsg = (f'{type(self)}.__add__(x) expects x of type {Colormap}; '
                      'but got x of type {type(other)}')
            raise TypeError(errmsg)
        # combine colors.
        #   use "shrink" to ensure colors have same format.
        #   E.g. without shrink (or some other evaluating of colors),
        #       might get shape (N, 3) instead of (N, 4).
        self_colors = self.shrink(0, 1).colors
        other_colors = other.shrink(0, 1).colors
        colors = list(self_colors) + list(other_colors)
        # combine names.
        self_name = '?' if self.name is None else self.name
        other_name = '?' if other.name is None else other.name
        name = f'{self_name}+{other_name}'
        # extremes
        kw_extremes = dict(under=self.under, over=other.over, bad=self.bad)
        # make result
        return self.get_cmap(colors=colors, name=name, **kw_extremes)


cmap = Colormap.get_cmap  # alias to Colormap.get_cmap
get_cmap = Colormap.get_cmap  # alias to Colormap.get_cmap

# # # some predefined cmaps # # #
CMAPS = {
    'growlight': cmap('viridis', under='#e8e8e8'),  # viridis with under = light gray.
    'growdark': cmap('viridis', under='#000028'),  # viridis with under = dark gray.
}
try:
    cet_fire = cmap('cet_fire')
except ValueError:  # didn't import colorcet --> default to 'inferno'.
    cet_fire = cmap('inferno')
else: # T = cet_fire with clear under (black) & non-white over (ivory)
    CMAPS['T'] = cet_fire.shrink(0.1, 0.95).with_over('ivory')
    del cet_fire  # remove from local namespace.



"""
File Purpose: misc. tools for reading directly from muram files
"""
import os
import re

import numpy as np

from ...errors import InputError


''' --------------------- muram param files --------------------- '''

def muram_snap_files(dir=os.curdir, *, abspath=False):
    '''return list of all muram Header.NNN files in directory.
    Sorts by snap number.

    abspath: bool. whether to return absolute paths or just the file basenames within directory.
    '''
    pattern = rf'Header[.]([0-9]+)'
    result = []
    for f in os.listdir(dir):
        match = re.fullmatch(pattern, f)
        if match is not None:
            snap_number = int(match.group(1))
            result.append((snap_number, f))
    result.sort()
    result = [f for (n, f) in result]
    if abspath:
        absdir = os.path.abspath(dir)
        result = [os.path.join(absdir, f) for f in result]
    return result

def read_muram_header(filename):
    '''Parse Muram Header.NNN file into a dictionary, containing:
        'order': 3-tuple containing 0,1,2 in same order as layout.order file
        'N0', 'N1', 'N2': number of cells in 0th, 1st, 2nd dimensions in data files.
        'Nx', 'Ny', 'Nz': number of cells in x, y, z dimensions
        'dx', 'dy', 'dz': cell width in each dimension
        't': time of this snapshot, in 'raw' units.
        'snap_s': snap number as a string, e.g. '004321'
        'dirname': directory containing the header file.

    Header file should contain at least 7 numbers, separated by whitespace:
        size0  size1  size2  ds0  ds1  ds2  time
    telling the number of cells (sizex) and cell width (dsx) in each dimension x=0,1,2,
        and the time t of this snapshot.

    params are mapped to x,y,z via the 'layout.order' file (in same dir as Header.NNN file).
        'layout.order' should contain only:
            0  1  2
        in any order, separated by whitespace.
    then e.g. the x size will be [size0, size1, size2][order[0]].
    (for y use order[1]; for z use order[2].)
    The dimensions in memmap files correspond to the order here;
        MuramCalculator.load_direct() handles transposing appropriately.
    '''
    filename = os.path.abspath(filename)
    if not os.path.isfile(filename):
        raise FileNotFoundError(f'file not found: {filename!r}')
    pattern = r'Header[.]([0-9]+)'
    match = re.fullmatch(pattern, os.path.basename(filename))
    if match is None:
        raise InputError(f'filename={os.path.basename(filename)!r} must match pattern={pattern!r}')
    snap_s = match.group(1)
    dirname = os.path.dirname(filename)
    layout_order_file = os.path.join(dirname, 'layout.order')
    if not os.path.isfile(layout_order_file):
        raise FileNotFoundError(f'layout.order file not found: {layout_order_file!r}')
    order = np.loadtxt(layout_order_file, dtype=int)
    header = np.loadtxt(filename)
    sizes_012 = header[:3].astype(int)
    sizes_xyz = sizes_012[order]
    ds = header[3:6][order]
    time = header[6]
    result = dict(order=order,
                  N0=sizes_012[0], N1=sizes_012[1], N2=sizes_012[2],
                  Nx=sizes_xyz[0], Ny=sizes_xyz[1], Nz=sizes_xyz[2],
                  dx=ds[0], dy=ds[1], dz=ds[2], t=time,
                  snap_s=snap_s, dirname=dirname,
                  )
    return result

def muram_directly_loadable_vars(snap_s, *, dir=os.curdir):
    '''return list of directly loadable vars for this snap, in this directory.
    snap_s: str
        snap number, as a string.
        E.g. if header file is 'Header.004321', snap_s should be '004321'.
    dir: str
        directory in which to search for files. Default: current directory.
    '''
    pattern = rf'(.+)[.]{snap_s}'  # var.{snap_s}
    result = []
    for f in os.listdir(dir):
        match = re.fullmatch(pattern, f)
        if match is not None:
            result.append(match.group(1))
    return sorted(result)



"""
File Purpose: calculating plasma heating, e.g. equilibrium T from E & collisions
"""

import numpy as np
import xarray as xr

from .plasma_parameters import PlasmaParametersLoader
from ..dimensions import CHARGED
from ..errors import QuantCalcError
from ..tools import xarray_sum


class PlasmaHeatingLoader(PlasmaParametersLoader):
    '''plasma heating. See help(self.get_Eheat) for more details.'''
    
    # # # HEATING OF CHARGED SPECIES # # #

    @known_var(deps=['m_n', 'skappa', 'mod_B', 'u_n'])
    def get_Eheat_perp_coeff(self, *, _Eheat_par_coeff=None):
        '''Eheat_perp = Eheat_perp_coeff * |E_perp|^2. for E heating perp to B. Units of Kelvin.
        see help(self.get_Eheat) for more details.

        [EFF] for efficiency, can provide _Eheat_par_coeff if known.
        '''
        Eheat_par_coeff = self('Eheat_par_coeff') if _Eheat_par_coeff is None else _Eheat_par_coeff
        return Eheat_par_coeff / (1 + self('skappa')**2)

    @known_var(deps=['m_n', 'skappa', 'mod_B', 'u_n'])
    def get_Eheat_par_coeff(self):
        '''Eheat_par = Eheat_par_coeff * |E_par|^2. for E heating parallel to B. Units of Kelvin.
        see help(self.get_Eheat) for more details.
        '''
        return (self('m_n') / (3 * self.u('kB'))) * (self('skappa')**2 / self('mod2_B'))

    @known_var(deps=['Eheat_perp_coeff', 'E_perpmag_B'])
    def get_Eheat_perp(self, *, _E_un0=None, _B=None, _Eheat_par_coeff=None):
        '''Eheat_perp = Eheat_perp_coeff * |E_perp|^2. heating perp to B. Units of Kelvin.
        see help(self.get_Eheat) for more details.

        [EFF] for efficiency, can provide _E_un0, _B and/or _Eheat_par_coeff if known.
            caution: if providing _E_un0 or _B, will assume any missing components are 0.
        '''
        perp_coeff = self('Eheat_perp_coeff', _Eheat_par_coeff=_Eheat_par_coeff)
        return perp_coeff * self('E_un0_perpmag_B', _E_un0=_E_un0, _B=_B)**2

    @known_var(deps=['Eheat_par_coeff', 'E_parmag_B'])
    def get_Eheat_par(self, *, _E_un0=None, _B=None, _Eheat_par_coeff=None):
        '''Eheat_par = Eheat_par_coeff * |E_par|^2. heating parallel to B. Units of Kelvin.
        see help(self.get_Eheat) for more details.

        [EFF] for efficiency, can provide _E_un0, _B and/or _Eheat_par_coeff if known.
            caution: if providing _E_un0 or _B, will assume any missing components are 0.
        '''
        par_coeff = self('Eheat_par_coeff') if _Eheat_par_coeff is None else _Eheat_par_coeff
        return par_coeff * self('E_un0_parmag_B', _E_un0=_E_un0, _B=_B)**2

    @known_var(deps=['Eheat_perp', 'Eheat_par'])
    def get_Eheat(self):
        '''Eheat = Eheat_perp + Eheat_par. total heating from electric field. Units of Kelvin.
        
        From assuming u_n=0 and derivatives=0 in heating & momentum equations, which yields:
            T_s = T_n + Eheat_perp + Eheat_par, where
                Eheat_perp = Eheat_perp_coeff * |E_perp|^2,
                Eheat_par  = Eheat_par_coeff * |E_par|^2,
                E_perp = E(in u_n=0 frame) perp to B,
                E_par  = E(in u_n=0 frame) parallel to B,
                Eheat_perp_coeff = (m_n / (3 kB)) (kappa_s^2 / B^2) * (1 / (1 + kappa_s^2)),
                Eheat_par_coeff  = (m_n / (3 kB)) (kappa_s^2 / B^2).
        '''
        with self.using(component=None):  # all 3 vector components
            E_un0 = self('E_un0')
            B = self('B')
            Eheat_par_coeff = self('Eheat_par_coeff')
        Eheat_perp = self('Eheat_perp', _E_un0=E_un0, _B=B, _Eheat_par_coeff=Eheat_par_coeff)
        Eheat_par = self('Eheat_par', _E_un0=E_un0, _B=B, _Eheat_par_coeff=Eheat_par_coeff)
        return Eheat_perp + Eheat_par


    # # # HEATING OF NEUTRALS # # #

    @known_var(deps=['Eheat_perp', 'nuns'])
    def get_Eheat_perp_rate_n_s(self):
        '''zeroth order rate of heating of neutrals due to collisions with s (self.fluid).
        Eheat_perp_rate_n_s = 2 * nuns * Eheat_perp
        == 2 * (m_s / m_n) * (n_s / n_n) * nusn * Eheat_perp
        == (2 m_s / (3 kB)) * (n_s / n_n) * nusn * (kappa_s^2 / (1 + kappa_s^2)) * (|E_perp|^2 / |B|^2)

        Derived from plugging T_from_Eheat_perp and |u_drift| formulae into the neutral heating equation:
            dTn/dt + (2/3) T_n div(u_n) =
                sum_s (2 m_n / (m_n + m_s)) nu_{n,s} [(m_s / (3 kB)) |u_s - u_n|^2 + (T_s - T_n)]
            in the neutral reference frame (u_n=0),
            and using m_n n_n nu_{n,s} = m_s n_s nu_{s,n} which comes from conservation of momentum.

        See also: Eheat_perp_rate_n
        '''
        Eheat_perp = self('Eheat_perp')  # on its own line for easy debugging in case of crash
        return 2 * self('nuns') * Eheat_perp

    @known_var(deps=['Eheat_perp_rate_n'], reduces_dims=['fluid'])
    def get_Eheat_perp_rate_n(self):
        '''zeroth order rate of heating of neutrals due to collisions with all charged species.
        Eheat_perp_rate_n = sum_s Eheat_perp_rate_n_s
        == sum_s 2 * nuns * Eheat_perp
        == sum_s 2 * (m_s / m_n) * (n_s / n_n) * nusn * Eheat_perp
        == sum_s (2 m_s / (3 kB)) * (n_s / n_n) * nusn * (kappa_s^2 / (1 + kappa_s^2)) * (|E_perp|^2 / |B|^2)
    
        Derived from plugging T_from_Eheat_perp and |u_drift| formulae into the neutral heating equation:
            dTn/dt + (2/3) T_n div(u_n) =
                sum_s (2 m_n / (m_n + m_s)) nu_{n,s} [(m_s / (3 kB)) |u_s - u_n|^2 + (T_s - T_n)]
            in the neutral reference frame (u_n=0),
            and using m_n n_n nu_{n,s} = m_s n_s nu_{s,n} which comes from conservation of momentum.
        '''
        return xarray_sum(self('Eheat_perp_rate_n_s', fluid=CHARGED), dim='fluid')

    @known_var(deps=['m', 'm_n', 'nuns', 'mod2_u', 'T', 'T_n'])
    def get_dTndt_s_ds(self):
        '''dataset of contributions to rate of heating of neutrals due to collisions with s.
        dTndt_s = 2 m_n / (m_n + m_s) * nuns * [(m_s / (3 kB)) |u_s|^2 + (T_s - T_n)]
        Assumes u_n==0; else crash with QuantCalcError.

        result has keys:
            'dTndt_u2': 2 m_n / (m_n + m_s) * nuns * [(m_s / (3 kB)) |u_s|^2]
            'dTndt_T': 2 m_n / (m_n + m_s) * nuns * [(T_s - T_n)]
        '''
        if not np.all(self('u_n')==0):
            raise QuantCalcError('dTndt implementation assumes u_n==0, but it is not.')
        outer = 2 * self('m_n/(m+m_n)') * self('nuns')
        inner_u = self('m/(3*kB)') * self('mod2_u')
        inner_T = self('T-T_n')
        result = xr.Dataset({
            'dTndt_u2': outer * inner_u,
            'dTndt_T': outer * inner_T,
        })
        return result

    @known_var(deps=['dTndt_s_ds'])
    def get_dTndt_s_u2(self):
        '''velocity contribution to rate of heating of neutrals due to collisions with s.
        dTndt_s_u2 = 2 m_n / (m_n + m_s) * nuns * [(m_s / (3 kB)) |u_s|^2]
        '''
        return self('dTndt_s_ds')['dTndt_u2']

    @known_var(deps=['dTndt_s_ds'])
    def get_dTndt_s_T(self):
        '''temperature contribution to rate of heating of neutrals due to collisions with s.
        dTndt_s_T = 2 m_n / (m_n + m_s) * nuns * [(T_s - T_n)]
        '''
        return self('dTndt_s_ds')['dTndt_T']

    @known_var(deps=['m', 'm_n', 'nuns', 'mod2_u', 'T', 'T_n'])
    def get_dTndt_s(self):
        '''rate of heating of neutrals due to collisions with s (self.fluid).
        dTndt_s = 2 m_n / (m_n + m_s) * nuns * [(m_s / (3 kB)) |u_s|^2 + (T_s - T_n)]
        '''
        ds = self('dTndt_s_ds')
        return ds['dTndt_u2'] + ds['dTndt_T']

    @known_var(deps=['dTndt_s'], ignores_dims=['fluid'])
    def get_dTndt(self):
        '''rate of heating of neutrals due to collisions with all charged species.
        dTndt = sum_s dTndt_s
        == sum_s 2 m_n / (m_n + m_s) * nuns * [(m_s / (3 kB)) |u_s|^2 + (T_s - T_n)]
        '''
        return xarray_sum(self('dTndt_s', fluid=CHARGED), dim='fluid')


    # # # PLASMA PARAMETERS AFFECTED BY HEATING # # #

    @known_var(deps=['Eheat', 'T_n'])
    def get_T_from_Eheat(self):
        '''T_from_Eheat = T_n + Eheat. Units of Kelvin.
        see help(self.get_Eheat) for more details.
        '''
        return self('T_n') + self('Eheat')

    @known_var(deps=['Eheat_perp', 'T_n'])
    def get_T_from_Eheat_perp(self):
        '''T_from_Eheat_perp = T_n + Eheat_perp. Units of Kelvin.
        see help(self.get_Eheat) for more details.
        '''
        return self('T_n') + self('Eheat_perp')

    @known_var(deps=['eqperp_ldebye2'])
    def get_eqperp_ldebye(self):
        '''Debye length (of self.fluid), using T_from_Eheat_perp instead of T.
        eqperp_ldebye = sqrt(epsilon0 kB T_from_Eheat_perp / (n q^2))
        '''
        return self('eqperp_ldebye2')**0.5

    @known_var(deps=['T_from_Eheat_perp', 'n', 'abs_q'])
    def get_eqperp_ldebye2(self):
        '''squared Debye length (of self.fluid), using Eheat_perp instead of T.
        eqperp_ldebye2 = epsilon0 kB T_from_Eheat_perp / (n q^2)
        '''
        T = self('T_from_Eheat_perp')
        return self.u('eps0') * self.u('kB') * T / (self('n') * self('abs_q')**2)

    @known_var(deps=['eqperp_ldebye2'], ignores_dims=['fluid'])
    def get_eqperp_ldebye_total(self):
        '''total Debye length for all fluids: sqrt(epsilon0 kB / sum_fluids(n q^2 / T)),
        using T = T_from_Eheat_perp.
        Equivalent: sqrt( 1 / sum_fluids(1/eqperp_ldebye^2) )
        '''
        return xarray_sum(1 / self('eqperp_ldebye2'), dim='fluid')**-0.5

    @known_var(deps=['T_from_Eheat_perp', 'm'])
    def get_eqperp_vtherm(self):
        '''thermal velocity, using T_from_Eheat_perp instead of T.
        eqperp_vtherm = sqrt(kB T_from_Eheat_perp / m)
        '''
        T = self('T_from_Eheat_perp')
        return (T * (self.u('kB') / self('m')))**0.5

    @known_var(deps=['dsmin_for_timescales', 'eqperp_vtherm'])
    def get_timescale_eqperp_vtherm(self):
        '''timescale from thermal velocity, using T_from_Eheat_perp instead of T.
        dsmin / eqperp_vtherm.
        '''
        return self('dsmin_for_timescales') / self('eqperp_vtherm')

    @known_var(deps=['eqperp_vtherm', 'nusn'], aliases=['eqperp_lmfp'])
    def get_eqperp_mean_free_path(self):
        '''collisional mean free path, using eqperp_vtherm instead of vtherm.
        eqperp_lmfp = eqperp_vtherm / nusn
        '''
        return self('eqperp_vtherm') / self('nusn')



"""
File Purpose: multifluid analysis from bifrost
"""

from .bifrost_multifluid_calculator import BifrostMultifluidCalculator
from .bifrost_multifluid_densities import BifrostMultifluidDensityLoader



"""
File Purpose: EppicPlotterManager which inherits from all the other EppicPlotterManagers here.
"""

from .eppic_plotters_sam import EppicPlotterManagerSam
from .eppic_plotters_save import EppicPlotterManagerSave

class EppicPlotterManager(EppicPlotterManagerSam, EppicPlotterManagerSave):
    '''PlotterManager inheriting from everyone's plotter manager codes here.'''
    # all functionality inherited from parents.
    pass



"""
File Purpose: basic test for bifrost hookup in PlasmaCalcs.
These tests require the snapname_NNN.idl file(s), but not any simulation output data.
"""
import os
import pytest

import numpy as np

import PlasmaCalcs as pc

HERE = os.path.dirname(__file__)
RUNDIR = os.path.join(HERE, 'test_bifrost_nodata')
SNAPNAME = 'l2d90x40r_it'  # snapname for the tests here.


''' --------------------- basics --------------------- '''

def get_bifrost_calculator(snapname=None, **kw_init):
    '''make a bifrost calculator'''
    with pc.InDir(RUNDIR):
        bc = pc.BifrostCalculator(snapname=snapname, **kw_init)
    return bc

def test_make_bifrost_calculator():
    '''ensure can make a bifrost calculator, when snapname is known.'''
    get_bifrost_calculator(snapname=SNAPNAME)


''' --------------------- inferring snapname --------------------- '''

def test_bifrost_infer_snapname():
    '''ensure can infer snapname from directory, when snapname is not provided.'''
    with pc.InDir(RUNDIR):
        snapname = pc.bifrost.bifrost_infer_snapname_here()
    assert snapname == SNAPNAME
    snapname = pc.bifrost.bifrost_infer_snapname_here(dir=RUNDIR)
    assert snapname == SNAPNAME
    # making bifrost calculator without supplying snapname only possible if snapname can be inferred.
    bc = get_bifrost_calculator(snapname=None)  # <-- just trying to ensure no crash.
    # ensure crash if snapname can't be inferred.
    #   (Ambiguity especially important to test, to avoid anyone accidentally reading the wrong data!)
    SNAPNAME_MISSING_DIR = os.path.join(HERE, 'test_infer_snapname', 'missing_crash')
    assert os.path.isdir(SNAPNAME_MISSING_DIR)
    with pytest.raises(FileNotFoundError):
        snapname = pc.bifrost.bifrost_infer_snapname_here(dir=SNAPNAME_MISSING_DIR)
    SNAPNAME_AMBIGUOUS_DIR = os.path.join(HERE, 'test_infer_snapname', 'ambiguous_crash')
    assert os.path.isdir(SNAPNAME_AMBIGUOUS_DIR)
    with pytest.raises(pc.FileAmbiguityError):
        snapname = pc.bifrost.bifrost_infer_snapname_here(dir=SNAPNAME_AMBIGUOUS_DIR)


''' --------------------- Units --------------------- '''

def test_bifrost_units():
    '''test bifrost units, also tests some of the cgs unit conversions code'''
    bc = get_bifrost_calculator()
    assert bc.u is not None
    assert isinstance(bc.u, pc.bifrost.BifrostUnitsManager)
    # ensure units kwarg is working
    bc = get_bifrost_calculator(units='raw')
    assert bc.u.units == bc.units == 'raw'
    bc = get_bifrost_calculator(units='si')
    assert bc.u.units == bc.units == 'si'
    bc = get_bifrost_calculator(units='cgs')  # ensure units='cgs' works too
    assert bc.u.units == bc.units == 'cgs'
    bc.units = 'si'
    assert bc.u.units == bc.units == 'si'
    # ensure default output is SI when .units == 'si'
    assert bc.u('l') == bc.u('l', 'si')
    assert bc.u('r') == bc.u('r', 'si')
    # try to get some other units just to ensure it's possible:
    for ustr in ('acceleration', 'current_density', 'temperature', 'pressure',
                 'flux', 'momentum_density', 'permittivity', 'capacitance',
                 'e_field', 'ef', 'b_field', 'b', 'force', 'power', 'volume',
                 'speed', 'q', 'charge', 'dimensionless', 'time', 'mass'):
        bc.u(ustr)
    # units test -- ensure can get conversion factors for all KNOWN:
    for key, ustr in bc.u.KNOWN.items():
        bc.u(ustr)
    # check units math works out properly:
    assert bc.u('l')**2 == bc.u('l2') == bc.u('area')
    assert bc.u('l')**-3 == bc.u('l-3') == bc.u('number_density') == bc.u('volume-1')
    assert bc.u('r')**-0.25 * bc.u('b')**3.7 == bc.u('r-0.25 b3.7')
    # checking units from params (bc.params tell conversions in cgs)
    assert bc.params is bc.params  # <-- ensure not recalculating params each time
    assert bc.u('l') * 1e2  == bc.params['u_l'] == bc.u('l', 'cgs')
    assert bc.u('t')        == bc.params['u_t'] == bc.u('t', 'cgs')
    assert bc.u('r') * 1e-3 == bc.params['u_r'] == bc.u('r', 'cgs')
    # check unit conversions:
    assert bc.u('l', 'si', convert_from='cgs') == 1e-2
    assert bc.u('l', 'cgs', 'si') == 1e2
    assert bc.u('M', 'cgs', 'si') == 1e3 == 1 / bc.u('M', 'si', 'cgs')
    assert bc.u('r', 'cgs', 'si') == 1e3 / (1e2)**3
    # check physical parameters:
    assert np.isclose(bc.u('amu', 'si'), bc.u.PHYSICAL_CONSTANTS_SI['amu'][0])
    assert np.isclose(bc.u('amu', 'cgs'), bc.u.PHYSICAL_CONSTANTS_SI['amu'][0] * 1e3)
    assert np.isclose(bc.u('c', 'si'), 2.99792E8)
    for pconst, (val, unit) in bc.u.PHYSICAL_CONSTANTS_SI.items():
        assert np.isclose(bc.u(pconst, 'si'), val)
    for pconst in ('amu', 'c', 'kB', 'me', 'h', 'eV', 'eV kB-1'):
        val, unit = bc.u.PHYSICAL_CONSTANTS_SI[pconst]
        assert np.isclose(bc.u(pconst, 'cgs'), val * bc.u(unit, 'cgs', 'si'))
    # for cgs, E & M quantities raise UnitsUnknownError:
    u = pc.UnitsManager()
    ucgs = u.CGS_UNITS
    for quant in ('q', 'b', 'ef', 'permittivity', 'permeability'):  # misc E & M conversion factors
        with pytest.raises(pc.UnitsUnknownError):
            ucgs(quant, 'si')  # converting from cgs to SI is impossible
        with pytest.raises(pc.UnitsUnknownError):
            u(quant, 'cgs')  # converting from raw to cgs is impossible
    for quant in ('eps0', 'mu0'):  # misc E & M physical constants
        with pytest.raises(pc.UnitsUnknownError):
            ucgs(quant, 'raw')  # converting to cgs is impossible
        with pytest.raises(pc.UnitsUnknownError):
            u(quant, 'cgs')  # converting to cgs is impossible


''' --------------------- Coords / Maindims / Meshfile --------------------- '''

def test_bifrost_load_meshfile():
    '''tests that we can load the meshfile coordinates!'''
    bc = get_bifrost_calculator()
    bc.load_mesh_coords()  # just checking that this doesn't crash
    bc.get_maindims_coords()  # just checking that this doesn't crash

def test_bifrost_squeeze_direct():
    '''tests that squeeze direct is working properly'''
    bc = get_bifrost_calculator()
    assert bc.params['mx'] > 1
    assert bc.params['my'] == 1  # in the input decks here, my=1.
    assert bc.params['mz'] > 1
    assert bc.maindims_size == bc.params['mx'] * bc.params['mz']
    # with squeezing:
    bc.squeeze_direct = True
    assert bc.maindims == ('x', 'z')
    assert bc.maindims_shape == (bc.params['mx'], bc.params['mz'])
    coords = bc.get_maindims_coords()
    assert 'y' not in coords
    assert 'x' in coords and 'z' in coords
    assert len(coords['x']) == bc.params['mx']
    assert len(coords['z']) == bc.params['mz']
    # without squeezing:
    bc.squeeze_direct = False
    assert bc.maindims == ('x', 'y', 'z')
    assert bc.maindims_shape == (bc.params['mx'], bc.params['my'], bc.params['mz'])
    coords = bc.get_maindims_coords()
    assert 'x' in coords and 'y' in coords and 'z' in coords
    assert len(coords['x']) == bc.params['mx']
    assert len(coords['y']) == bc.params['my']
    assert len(coords['z']) == bc.params['mz']

# [TODO] could test stagger methods here if desired.
